# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from __future__ import annotations

import base64
import hashlib
from typing import TYPE_CHECKING
from typing import Any
from urllib.parse import urlencode
from urllib.parse import urljoin
from urllib.parse import urlparse
from urllib.parse import urlunparse

from twisted.internet import defer
from twisted.python import log

from buildbot import config
from buildbot.util import bytes2unicode
from buildbot.util import httpclientservice
from buildbot.util import unicode2bytes
from buildbot.util.config import ConfiguredMixin
from buildbot.www import resource

if TYPE_CHECKING:
    from buildbot.master import BuildMaster
    from buildbot.util.twisted import InlineCallbacksType


class AvatarBase(ConfiguredMixin):
    name = "noavatar"

    master: BuildMaster | None = None

    def getUserAvatar(
        self, email: bytes, username: bytes | None, size: int, defaultAvatarUrl: str
    ) -> Any:
        raise NotImplementedError()


class AvatarGitHub(AvatarBase):
    name = "github"

    DEFAULT_GITHUB_API_URL = 'https://api.github.com'

    client: httpclientservice.HTTPSession | None = None

    def __init__(
        self,
        github_api_endpoint: str | None = None,
        token: str | None = None,
        client_id: str | None = None,
        client_secret: str | None = None,
        debug: bool = False,
        verify: bool = True,
    ) -> None:
        self.github_api_endpoint = github_api_endpoint
        if github_api_endpoint is None:
            self.github_api_endpoint = self.DEFAULT_GITHUB_API_URL
        self.token = token
        self.client_creds = None
        if bool(client_id) != bool(client_secret):
            config.error('client_id and client_secret must be both provided or none')
        if client_id:
            if token:
                config.error('client_id and client_secret must not be provided when token is')
            # We know client_id and client_secret are not None here because of the check above
            self.client_creds = base64.b64encode(
                b':'.join(cred.encode('utf-8') for cred in (client_id, client_secret))  # type: ignore[union-attr]
            ).decode('ascii')
        self.debug = debug
        self.verify = verify

        self.master = None
        self.client = None

    @defer.inlineCallbacks
    def _get_http_client(self) -> InlineCallbacksType[httpclientservice.HTTPSession]:
        assert self.master is not None

        if self.client is not None:
            return self.client

        headers = {
            'User-Agent': 'Buildbot',
        }
        if self.token:
            headers['Authorization'] = 'token ' + self.token
        elif self.client_creds:
            headers['Authorization'] = 'basic ' + self.client_creds

        self.client = yield httpclientservice.HTTPSession(
            self.master.httpservice,
            self.github_api_endpoint,
            headers=headers,
            debug=self.debug,
            verify=self.verify,
        )

        return self.client

    @defer.inlineCallbacks
    def _get_avatar_by_username(self, username: str) -> InlineCallbacksType[str | None]:
        headers = {
            'Accept': 'application/vnd.github.v3+json',
        }

        url = f'/users/{username}'
        http = yield self._get_http_client()
        res = yield http.get(url, headers=headers)
        if res.code == 404:
            # Not found
            return None
        if 200 <= res.code < 300:
            data = yield res.json()
            return data['avatar_url']

        log.msg(f'Failed looking up user: response code {res.code}')
        return None

    @defer.inlineCallbacks
    def _search_avatar_by_user_email(self, email: str) -> InlineCallbacksType[str | None]:
        headers = {
            'Accept': 'application/vnd.github.v3+json',
        }

        query = f'{email} in:email'
        url = f"/search/users?{urlencode({'q': query})}"
        http = yield self._get_http_client()
        res = yield http.get(url, headers=headers)
        if 200 <= res.code < 300:
            data = yield res.json()
            if data['total_count'] == 0:
                # Not found
                return None
            return data['items'][0]['avatar_url']

        log.msg(f'Failed searching user by email: response code {res.code}')
        return None

    @defer.inlineCallbacks
    def _search_avatar_by_commit(self, email: str) -> InlineCallbacksType[str | None]:
        headers = {
            'Accept': 'application/vnd.github.v3+json,application/vnd.github.cloak-preview',
        }

        query = {
            'q': f'author-email:{email}',
            'sort': 'committer-date',
            'per_page': '1',
        }
        sorted_query = sorted(query.items(), key=lambda x: x[0])
        url = f'/search/commits?{urlencode(sorted_query)}'
        http = yield self._get_http_client()
        res = yield http.get(url, headers=headers)
        if 200 <= res.code < 300:
            data = yield res.json()
            if data['total_count'] == 0:
                # Not found
                return None
            author = data['items'][0]['author']
            if author is None:
                # No Github account found
                return None
            return author['avatar_url']

        log.msg(f'Failed searching user by commit: response code {res.code}')
        return None

    def _add_size_to_url(self, avatar: str, size: int) -> str:
        parts = urlparse(avatar)
        query = parts.query
        if query:
            query += '&'
        query += f's={size}'
        return urlunparse((
            parts.scheme,
            parts.netloc,
            parts.path,
            parts.params,
            query,
            parts.fragment,
        ))

    @defer.inlineCallbacks
    def getUserAvatar(
        self, email: bytes, username: bytes | None, size: int, defaultAvatarUrl: str
    ) -> InlineCallbacksType[Any | None]:
        username_str = username.decode('utf-8') if username else None
        email_str = email.decode('utf-8') if email else None

        avatar: str | None = None
        if username_str:
            avatar = yield self._get_avatar_by_username(username_str)
        if not avatar and email_str:
            avatar = yield self._search_avatar_by_user_email(email_str)
        if not avatar and email_str:
            # No luck, try to find a commit with this email
            avatar = yield self._search_avatar_by_commit(email_str)

        if not avatar:
            # No luck
            return None

        if size:
            avatar = self._add_size_to_url(avatar, size)

        raise resource.Redirect(avatar)


class AvatarGravatar(AvatarBase):
    name = "gravatar"
    # gravatar does not want intranet URL, which is most of where the bots are
    # just use same default as github (retro)
    default = "retro"

    def getUserAvatar(
        self, email: bytes, username: bytes | None, size: int, defaultAvatarUrl: str
    ) -> Any:
        # construct the url
        emailBytes = unicode2bytes(email.lower())
        emailHash = hashlib.md5(emailBytes)
        gravatar_url = "//www.gravatar.com/avatar/"
        gravatar_url += emailHash.hexdigest() + "?"
        if self.default != "url":  # type: ignore[comparison-overlap]
            defaultAvatarUrl = self.default
        url = {'d': defaultAvatarUrl, 's': str(size)}
        sorted_url = sorted(url.items(), key=lambda x: x[0])
        gravatar_url += urlencode(sorted_url)
        raise resource.Redirect(gravatar_url)


class AvatarResource(resource.Resource):
    # enable reconfigResource calls
    needsReconfig = True
    defaultAvatarUrl = b"img/nobody.png"

    avatarMethods: list[AvatarBase] = []
    defaultAvatarFullUrl: bytes
    cache: dict[tuple[bytes, bytes | None, int], Any]

    def reconfigResource(self, new_config: Any) -> None:
        avatar_methods = new_config.www.get('avatar_methods', [])
        self.defaultAvatarFullUrl = urljoin(
            unicode2bytes(new_config.buildbotURL), unicode2bytes(self.defaultAvatarUrl)
        )
        self.cache = {}

        # ensure the avatarMethods is a iterable
        if isinstance(avatar_methods, AvatarBase):
            self.avatarMethods = [avatar_methods]
        else:
            self.avatarMethods = avatar_methods

        for method in self.avatarMethods:
            method.master = self.master

    def render_GET(self, request: Any) -> Any:
        return self.asyncRenderHelper(request, self.renderAvatar)

    @defer.inlineCallbacks
    def renderAvatar(self, request: Any) -> InlineCallbacksType[Any]:
        email = request.args.get(b"email", [b""])[0]
        size = request.args.get(b"size", [32])[0]
        try:
            size = int(size)
        except ValueError:
            size = 32
        username = request.args.get(b"username", [None])[0]
        cache_key = (email, username, size)
        if self.cache.get(cache_key):
            raise self.cache[cache_key]
        for method in self.avatarMethods:
            try:
                res = yield method.getUserAvatar(
                    email, username, size, bytes2unicode(self.defaultAvatarFullUrl)
                )
            except resource.Redirect as r:
                self.cache[cache_key] = r
                raise
            if res is not None:
                request.setHeader(b'content-type', res[0])
                request.setHeader(b'content-length', unicode2bytes(str(len(res[1]))))
                request.write(res[1])
                return
        raise resource.Redirect(self.defaultAvatarUrl)



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from __future__ import annotations

from typing import TYPE_CHECKING

from twisted.internet import defer

if TYPE_CHECKING:
    from typing import Any

    from twisted.trial import unittest

    from buildbot.test.fake import fakemaster
    from buildbot.util.twisted import InlineCallbacksType

    _StateTestMixinBase = unittest.TestCase
else:
    _StateTestMixinBase = object


class StateTestMixin(_StateTestMixinBase):
    master: fakemaster.FakeMaster

    @defer.inlineCallbacks
    def set_fake_state(self, object: Any, name: str, value: Any) -> InlineCallbacksType[None]:
        objectid = yield self.master.db.state.getObjectId(object.name, object.__class__.__name__)
        yield self.master.db.state.setState(objectid, name, value)

    @defer.inlineCallbacks
    def assert_state(self, objectid: int, **kwargs: Any) -> InlineCallbacksType[None]:
        for k, v in kwargs.items():
            value = yield self.master.db.state.getState(objectid, k)
            self.assertEqual(value, v, f"state for {k!r} is {v!r}")

    @defer.inlineCallbacks
    def assert_state_by_class(
        self, name: str, class_name: str, **kwargs: Any
    ) -> InlineCallbacksType[None]:
        objectid = yield self.master.db.state.getObjectId(name, class_name)
        yield self.assert_state(objectid, **kwargs)



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

#
# Parts of this code were copied from Twisted Python.
# Copyright (c) Twisted Matrix Laboratories.
#
from __future__ import annotations

from typing import TYPE_CHECKING

from twisted.internet import defer
from twisted.internet import interfaces
from twisted.internet import protocol
from zope.interface import implementer

if TYPE_CHECKING:
    from twisted.internet.interfaces import IAddress
    from twisted.internet.interfaces import IConnector
    from twisted.internet.interfaces import IProtocol
    from twisted.internet.interfaces import IProtocolFactory
    from twisted.internet.interfaces import IStreamClientEndpoint
    from twisted.internet.protocol import Protocol
    from twisted.python.failure import Failure


class HTTPTunnelClient(protocol.Protocol):
    """
    This protocol handles the HTTP communication with the proxy server
    and subsequent creation of the tunnel.

    Once the tunnel is established, all incoming communication is forwarded
    directly to the wrapped protocol.
    """

    def __init__(self, connectedDeferred: defer.Deferred[IProtocol]) -> None:
        # this gets set once the tunnel is ready
        self._proxyWrappedProtocol: IProtocol | None = None
        self._connectedDeferred = connectedDeferred

    def connectionMade(self) -> None:
        assert isinstance(self.factory, HTTPTunnelFactory)
        request = f"CONNECT {self.factory.host}:{self.factory.port} HTTP/1.1\r\n\r\n"
        assert self.transport is not None
        self.transport.write(request.encode())

    def connectionLost(self, reason: Failure = protocol.connectionDone) -> None:
        if self._proxyWrappedProtocol:
            # Proxy connectionLost to the wrapped protocol
            self._proxyWrappedProtocol.connectionLost(reason)

    def dataReceived(self, data: bytes) -> None:
        if self._proxyWrappedProtocol is not None:
            # If tunnel is already established, proxy dataReceived()
            # calls to the wrapped protocol
            return self._proxyWrappedProtocol.dataReceived(data)

        assert self.transport is not None

        # process data from the proxy server
        _, status, _ = data.split(b"\r\n")[0].split(b" ", 2)
        if status != b"200":
            self.transport.loseConnection()
            return

        assert isinstance(self.factory, HTTPTunnelFactory)
        self._proxyWrappedProtocol = self.factory._proxyWrappedFactory.buildProtocol(
            self.transport.getPeer()
        )
        assert self._proxyWrappedProtocol is not None
        self._proxyWrappedProtocol.makeConnection(self.transport)
        self._connectedDeferred.callback(self._proxyWrappedProtocol)

        # forward all traffic directly to the wrapped protocol
        self.transport.protocol = self._proxyWrappedProtocol  # type: ignore[attr-defined]

        # In case the server sent some data together with its response,
        # forward those to the wrapped protocol.
        remaining_data = data.split(b"\r\n\r\n", 2)[1]
        if remaining_data:
            self._proxyWrappedProtocol.dataReceived(remaining_data)
            return

        return None


class HTTPTunnelFactory(protocol.ClientFactory):
    """The protocol factory for the HTTP tunnel.

    It is used as a wrapper for BotFactory, which can hence be shielded
    from all the proxy business.
    """

    protocol = HTTPTunnelClient  # type: ignore[assignment]

    def __init__(self, host: str, port: int, wrappedFactory: IProtocolFactory) -> None:
        self.host = host
        self.port = port

        self._proxyWrappedFactory = wrappedFactory
        self._onConnection: defer.Deferred[IProtocol] = defer.Deferred()

    def doStart(self) -> None:
        super().doStart()
        # forward start notifications through to the wrapped factory.
        self._proxyWrappedFactory.doStart()

    def doStop(self) -> None:
        # forward stop notifications through to the wrapped factory.
        self._proxyWrappedFactory.doStop()
        super().doStop()

    def buildProtocol(self, addr: IAddress) -> Protocol | None:
        proto = self.protocol(self._onConnection)  # type: ignore[has-type]
        proto.factory = self
        return proto

    def clientConnectionFailed(
        self,
        connector: IConnector,
        reason: Failure | BaseException | None,
    ) -> None:
        if not self._onConnection.called:
            self._onConnection.errback(reason)


@implementer(interfaces.IStreamClientEndpoint)
class HTTPTunnelEndpoint:
    """This handles the connection to buildbot master on given 'host'
    and 'port' through the proxy server given as 'proxyEndpoint'.
    """

    def __init__(self, host: str, port: int, proxyEndpoint: IStreamClientEndpoint) -> None:
        self.host = host
        self.port = port
        self.proxyEndpoint = proxyEndpoint

    def connect(self, protocolFactory: IProtocolFactory) -> defer.Deferred:
        """Connect to remote server through an HTTP tunnel."""
        tunnel = HTTPTunnelFactory(self.host, self.port, protocolFactory)
        d = self.proxyEndpoint.connect(tunnel)
        # once tunnel connection is established,
        # defer the subsequent server connection
        d.addCallback(lambda result: tunnel._onConnection)
        return d



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members
from __future__ import annotations

from unittest.mock import Mock

from twisted.internet import defer
from twisted.internet import reactor
from twisted.python import failure
from twisted.trial import unittest
from zope.interface import implementer

from buildbot import config
from buildbot import interfaces
from buildbot.process import properties
from buildbot.process.results import CANCELLED
from buildbot.process.results import EXCEPTION
from buildbot.process.results import FAILURE
from buildbot.process.results import SUCCESS
from buildbot.steps import trigger
from buildbot.test import fakedb
from buildbot.test.reactor import TestReactorMixin
from buildbot.test.steps import TestBuildStepMixin
from buildbot.test.util.interfaces import InterfaceTests


@implementer(interfaces.ITriggerableScheduler)
class FakeTriggerable:
    triggered_with = None
    result = SUCCESS
    bsid = 1
    brids: dict[int, int] = {}
    exception = False
    never_finish = False

    def __init__(self, name):
        self.name = name

    def trigger(
        self,
        waited_for,
        sourcestamps=None,
        set_props=None,
        parent_buildid=None,
        parent_relationship=None,
    ):
        self.triggered_with = (waited_for, sourcestamps, set_props.properties)
        idsDeferred = defer.Deferred()
        idsDeferred.callback((self.bsid, self.brids))
        resultsDeferred = defer.Deferred()
        if not self.never_finish:
            if self.exception:
                reactor.callLater(0, resultsDeferred.errback, RuntimeError('oh noes'))
            else:
                reactor.callLater(0, resultsDeferred.callback, (self.result, self.brids))
        return (idsDeferred, resultsDeferred)


class TriggerableInterfaceTest(unittest.TestCase, InterfaceTests):
    def test_interface(self):
        self.assertInterfacesImplemented(FakeTriggerable)


class FakeSourceStamp:
    def __init__(self, **kwargs):
        self.__dict__.update(kwargs)

    def asDict(self, includePatch=True):
        return self.__dict__.copy()


class FakeSchedulerManager:
    pass


# Magic numbers that relate brid to other build settings
def BRID_TO_BSID(brid):
    return brid + 2000


def BRID_TO_BID(brid):
    return brid + 3000


def BRID_TO_BUILD_NUMBER(brid):
    return brid + 4000


class TestTrigger(TestBuildStepMixin, TestReactorMixin, unittest.TestCase):
    def setUp(self):
        self.setup_test_reactor()
        return self.setup_test_build_step()

    @defer.inlineCallbacks
    def setup_step(self, step, sourcestampsInBuild=None, gotRevisionsInBuild=None, *args, **kwargs):
        sourcestamps = sourcestampsInBuild or []
        got_revisions = gotRevisionsInBuild or {}

        yield super().setup_step(step, *args, **kwargs)

        # This step reaches deeply into a number of parts of Buildbot.  That
        # should be fixed!

        # set up a buildmaster that knows about two fake schedulers, a and b
        m = self.master
        self.build.builder.botmaster = m.botmaster
        self.build.conn = object()
        m.config.buildbotURL = "baseurl/"
        m.scheduler_manager = FakeSchedulerManager()

        self.scheduler_a = a = FakeTriggerable(name='a')
        self.scheduler_b = b = FakeTriggerable(name='b')
        self.scheduler_c = c = FakeTriggerable(name='c')
        m.scheduler_manager.namedServices = {"a": a, "b": b, "c": c}

        a.brids = {77: 11}
        b.brids = {78: 22}
        c.brids = {79: 33, 80: 44}

        def make_fake_br(brid, builderid):
            return fakedb.BuildRequest(id=brid, buildsetid=BRID_TO_BSID(brid), builderid=builderid)

        def make_fake_build(brid, builderid):
            return fakedb.Build(
                buildrequestid=brid,
                id=BRID_TO_BID(brid),
                number=BRID_TO_BUILD_NUMBER(brid),
                masterid=9,
                workerid=13,
                builderid=builderid,
            )

        yield m.db.insert_test_data([
            fakedb.Builder(id=77, name='A'),
            fakedb.Builder(id=78, name='B'),
            fakedb.Builder(id=79, name='C1'),
            fakedb.Builder(id=80, name='C2'),
            fakedb.Master(id=9),
            fakedb.Buildset(id=2022),
            fakedb.Buildset(id=2011),
            fakedb.Buildset(id=2033),
            fakedb.Worker(id=13, name="some:worker"),
            make_fake_br(11, 77),
            make_fake_br(22, 78),
            fakedb.BuildRequest(id=33, buildsetid=2033, builderid=79),
            fakedb.BuildRequest(id=44, buildsetid=2033, builderid=80),
            make_fake_build(11, builderid=77),
            make_fake_build(22, builderid=78),
            make_fake_build(33, builderid=79),
            # builderid is 79 on purpose, changed, from the one of the buildrequest
            # to test the case of the virtual
            make_fake_build(44, builderid=79),
        ])

        def getAllSourceStamps():
            return sourcestamps

        self.build.getAllSourceStamps = getAllSourceStamps

        def getAllGotRevisions():
            return got_revisions

        self.get_nth_step(0).getAllGotRevisions = getAllGotRevisions

        self.exp_add_sourcestamp = None
        self.exp_a_trigger = None
        self.exp_b_trigger = None
        self.exp_c_trigger = None
        self.exp_added_urls = []

    @defer.inlineCallbacks
    def run_step(self, results_dict=None):
        if results_dict is None:
            results_dict = {}
        if self.get_nth_step(0).waitForFinish:
            for i in [11, 22, 33, 44]:
                yield self.master.db.builds.finishBuild(
                    BRID_TO_BID(i), results_dict.get(i, SUCCESS)
                )
        d = super().run_step()
        # the build doesn't finish until after a callLater, so this has the
        # effect of checking whether the deferred has been fired already;
        if self.get_nth_step(0).waitForFinish:
            self.assertFalse(d.called)
        else:
            self.assertTrue(d.called)

        yield d
        self.assertEqual(self.scheduler_a.triggered_with, self.exp_a_trigger)
        self.assertEqual(self.scheduler_b.triggered_with, self.exp_b_trigger)

        # check the URLs
        step_data = yield self.master.data.get(('steps', self.get_nth_step(0).stepid))
        step_urls = [(url['name'], url['url']) for url in step_data['urls']]
        self.assertEqual(step_urls, sorted(self.exp_added_urls))

        if self.exp_add_sourcestamp:
            self.assertEqual(self.addSourceStamp_kwargs, self.exp_add_sourcestamp)

        # pause run_step's completion until after any other callLater's are done
        d = defer.Deferred()
        reactor.callLater(0, d.callback, None)
        yield d

    def expectTriggeredWith(self, a=None, b=None, c=None, d=None):
        self.exp_a_trigger = a
        if a is not None:
            self.expectTriggeredLinks('a_br')
        self.exp_b_trigger = b
        if b is not None:
            self.expectTriggeredLinks('b_br')
        self.exp_c_trigger = c
        if c is not None:
            self.expectTriggeredLinks('c_br')

    def expectAddedSourceStamp(self, **kwargs):
        self.exp_add_sourcestamp = kwargs

    def expectTriggeredLinks(self, *args):
        if 'a_br' in args:
            self.exp_added_urls.append(('a #11', 'baseurl/#/buildrequests/11'))
        if 'b_br' in args:
            self.exp_added_urls.append(('b #22', 'baseurl/#/buildrequests/22'))
        if 'c_br' in args:
            self.exp_added_urls.append(('c #33', 'baseurl/#/buildrequests/33'))
            self.exp_added_urls.append(('c #44', 'baseurl/#/buildrequests/44'))
        if 'a' in args:
            self.exp_added_urls.append(('success: A #4011', 'baseurl/#/builders/77/builds/4011'))
        if 'b' in args:
            self.exp_added_urls.append(('success: B #4022', 'baseurl/#/builders/78/builds/4022'))
        if 'afailed' in args:
            self.exp_added_urls.append(('failure: A #4011', 'baseurl/#/builders/77/builds/4011'))
        if 'c' in args:
            self.exp_added_urls.append(('success: C1 #4033', 'baseurl/#/builders/79/builds/4033'))
            self.exp_added_urls.append(('success: C1 #4044', 'baseurl/#/builders/79/builds/4044'))

    # tests
    def test_no_schedulerNames(self):
        with self.assertRaises(config.ConfigErrors):
            trigger.Trigger()

    def test_unimportantSchedulerNames_not_in_schedulerNames(self):
        with self.assertRaises(config.ConfigErrors):
            trigger.Trigger(schedulerNames=['a'], unimportantSchedulerNames=['b'])

    def test_unimportantSchedulerNames_not_in_schedulerNames_but_rendered(self):
        # should not raise
        trigger.Trigger(
            schedulerNames=[properties.Interpolate('a')], unimportantSchedulerNames=['b']
        )

    def test_sourceStamp_and_updateSourceStamp(self):
        with self.assertRaises(config.ConfigErrors):
            trigger.Trigger(schedulerNames=['c'], sourceStamp={"x": 1}, updateSourceStamp=True)

    def test_sourceStamps_and_updateSourceStamp(self):
        with self.assertRaises(config.ConfigErrors):
            trigger.Trigger(
                schedulerNames=['c'], sourceStamps=[{"x": 1}, {"x": 2}], updateSourceStamp=True
            )

    def test_updateSourceStamp_and_alwaysUseLatest(self):
        with self.assertRaises(config.ConfigErrors):
            trigger.Trigger(schedulerNames=['c'], updateSourceStamp=True, alwaysUseLatest=True)

    def test_sourceStamp_and_alwaysUseLatest(self):
        with self.assertRaises(config.ConfigErrors):
            trigger.Trigger(schedulerNames=['c'], sourceStamp={"x": 1}, alwaysUseLatest=True)

    def test_sourceStamps_and_alwaysUseLatest(self):
        with self.assertRaises(config.ConfigErrors):
            trigger.Trigger(
                schedulerNames=['c'], sourceStamps=[{"x": 1}, {"x": 2}], alwaysUseLatest=True
            )

    @defer.inlineCallbacks
    def test_simple(self):
        yield self.setup_step(trigger.Trigger(schedulerNames=['a'], sourceStamps={}))
        self.expect_outcome(result=SUCCESS, state_string='triggered a')
        self.expectTriggeredWith(a=(False, [], {}))
        yield self.run_step()

    @defer.inlineCallbacks
    def test_simple_failure(self):
        yield self.setup_step(trigger.Trigger(schedulerNames=['a']))
        self.scheduler_a.result = FAILURE
        # not waitForFinish, so trigger step succeeds even though the build
        # didn't fail
        self.expect_outcome(result=SUCCESS, state_string='triggered a')
        self.expectTriggeredWith(a=(False, [], {}))
        yield self.run_step()

    @defer.inlineCallbacks
    def test_simple_exception(self):
        yield self.setup_step(trigger.Trigger(schedulerNames=['a']))
        self.scheduler_a.exception = True
        self.expect_outcome(result=SUCCESS, state_string='triggered a')
        self.expectTriggeredWith(a=(False, [], {}))
        yield self.run_step()

        self.assertEqual(len(self.flushLoggedErrors(RuntimeError)), 1)

    @defer.inlineCallbacks
    def test_bogus_scheduler(self):
        yield self.setup_step(trigger.Trigger(schedulerNames=['a', 'x']))
        # bogus scheduler is an exception, not a failure (don't blame the patch)
        self.expect_outcome(result=EXCEPTION)
        self.expectTriggeredWith(a=None)  # a is not triggered!
        yield self.run_step()
        self.flushLoggedErrors(ValueError)

    @defer.inlineCallbacks
    def test_updateSourceStamp(self):
        yield self.setup_step(
            trigger.Trigger(schedulerNames=['a'], updateSourceStamp=True),
            sourcestampsInBuild=[FakeSourceStamp(codebase='', repository='x', revision=11111)],
            gotRevisionsInBuild={'': 23456},
        )
        self.expect_outcome(result=SUCCESS, state_string='triggered a')
        self.expectTriggeredWith(
            a=(False, [{'codebase': '', 'repository': 'x', 'revision': 23456}], {})
        )
        yield self.run_step()

    @defer.inlineCallbacks
    def test_updateSourceStamp_no_got_revision(self):
        yield self.setup_step(
            trigger.Trigger(schedulerNames=['a'], updateSourceStamp=True),
            sourcestampsInBuild=[FakeSourceStamp(codebase='', repository='x', revision=11111)],
        )
        self.expect_outcome(result=SUCCESS)
        self.expectTriggeredWith(
            a=(
                False,
                # uses old revision
                [{'codebase': '', 'repository': 'x', 'revision': 11111}],
                {},
            )
        )
        yield self.run_step()

    @defer.inlineCallbacks
    def test_not_updateSourceStamp(self):
        yield self.setup_step(
            trigger.Trigger(schedulerNames=['a'], updateSourceStamp=False),
            sourcestampsInBuild=[FakeSourceStamp(codebase='', repository='x', revision=11111)],
            gotRevisionsInBuild={'': 23456},
        )
        self.expect_outcome(result=SUCCESS)
        self.expectTriggeredWith(
            a=(False, [{'codebase': '', 'repository': 'x', 'revision': 11111}], {})
        )
        yield self.run_step()

    @defer.inlineCallbacks
    def test_updateSourceStamp_multiple_repositories(self):
        yield self.setup_step(
            trigger.Trigger(schedulerNames=['a'], updateSourceStamp=True),
            sourcestampsInBuild=[
                FakeSourceStamp(codebase='cb1', revision='12345'),
                FakeSourceStamp(codebase='cb2', revision='12345'),
            ],
            gotRevisionsInBuild={'cb1': 23456, 'cb2': 34567},
        )
        self.expect_outcome(result=SUCCESS)
        self.expectTriggeredWith(
            a=(
                False,
                [{'codebase': 'cb1', 'revision': 23456}, {'codebase': 'cb2', 'revision': 34567}],
                {},
            )
        )
        yield self.run_step()

    @defer.inlineCallbacks
    def test_updateSourceStamp_prop_false(self):
        yield self.setup_step(
            trigger.Trigger(schedulerNames=['a'], updateSourceStamp=properties.Property('usess')),
            sourcestampsInBuild=[FakeSourceStamp(codebase='', repository='x', revision=11111)],
            gotRevisionsInBuild={'': 23456},
        )
        self.build.setProperty('usess', False, 'me')
        self.expect_outcome(result=SUCCESS)
        # didn't use got_revision
        self.expectTriggeredWith(
            a=(False, [{'codebase': '', 'repository': 'x', 'revision': 11111}], {})
        )
        yield self.run_step()

    @defer.inlineCallbacks
    def test_updateSourceStamp_prop_true(self):
        yield self.setup_step(
            trigger.Trigger(schedulerNames=['a'], updateSourceStamp=properties.Property('usess')),
            sourcestampsInBuild=[FakeSourceStamp(codebase='', repository='x', revision=11111)],
            gotRevisionsInBuild={'': 23456},
        )
        self.build.setProperty('usess', True, 'me')
        self.expect_outcome(result=SUCCESS)
        # didn't use got_revision
        self.expectTriggeredWith(
            a=(False, [{'codebase': '', 'repository': 'x', 'revision': 23456}], {})
        )
        yield self.run_step()

    @defer.inlineCallbacks
    def test_alwaysUseLatest(self):
        yield self.setup_step(
            trigger.Trigger(schedulerNames=['b'], alwaysUseLatest=True),
            sourcestampsInBuild=[FakeSourceStamp(codebase='', repository='x', revision=11111)],
        )
        self.expect_outcome(result=SUCCESS)
        # Do not pass setid
        self.expectTriggeredWith(b=(False, [], {}))
        yield self.run_step()

    @defer.inlineCallbacks
    def test_alwaysUseLatest_prop_false(self):
        yield self.setup_step(
            trigger.Trigger(schedulerNames=['b'], alwaysUseLatest=properties.Property('aul')),
            sourcestampsInBuild=[FakeSourceStamp(codebase='', repository='x', revision=11111)],
        )
        self.build.setProperty('aul', False, 'me')
        self.expect_outcome(result=SUCCESS)
        # didn't use latest
        self.expectTriggeredWith(
            b=(False, [{'codebase': '', 'repository': 'x', 'revision': 11111}], {})
        )
        yield self.run_step()

    @defer.inlineCallbacks
    def test_alwaysUseLatest_prop_true(self):
        yield self.setup_step(
            trigger.Trigger(schedulerNames=['b'], alwaysUseLatest=properties.Property('aul')),
            sourcestampsInBuild=[FakeSourceStamp(codebase='', repository='x', revision=11111)],
        )
        self.build.setProperty('aul', True, 'me')
        self.expect_outcome(result=SUCCESS)
        # didn't use latest
        self.expectTriggeredWith(b=(False, [], {}))
        yield self.run_step()

    @defer.inlineCallbacks
    def test_sourceStamp(self):
        ss = {"revision": 9876, "branch": 'dev'}
        yield self.setup_step(trigger.Trigger(schedulerNames=['b'], sourceStamp=ss))
        self.expect_outcome(result=SUCCESS)
        self.expectTriggeredWith(b=(False, [ss], {}))
        yield self.run_step()

    @defer.inlineCallbacks
    def test_set_of_sourceStamps(self):
        ss1 = {"codebase": 'cb1', "repository": 'r1', "revision": 9876, "branch": 'dev'}
        ss2 = {"codebase": 'cb2', "repository": 'r2', "revision": 5432, "branch": 'dev'}
        yield self.setup_step(trigger.Trigger(schedulerNames=['b'], sourceStamps=[ss1, ss2]))
        self.expect_outcome(result=SUCCESS)
        self.expectTriggeredWith(b=(False, [ss1, ss2], {}))
        yield self.run_step()

    @defer.inlineCallbacks
    def test_set_of_sourceStamps_override_build(self):
        ss1 = {"codebase": 'cb1', "repository": 'r1', "revision": 9876, "branch": 'dev'}
        ss2 = {"codebase": 'cb2', "repository": 'r2', "revision": 5432, "branch": 'dev'}
        ss3 = FakeSourceStamp(codebase='cb3', repository='r3', revision=1234, branch='dev')
        ss4 = FakeSourceStamp(codebase='cb4', repository='r4', revision=2345, branch='dev')
        yield self.setup_step(
            trigger.Trigger(schedulerNames=['b'], sourceStamps=[ss1, ss2]),
            sourcestampsInBuild=[ss3, ss4],
        )
        self.expect_outcome(result=SUCCESS)
        self.expectTriggeredWith(b=(False, [ss1, ss2], {}))
        yield self.run_step()

    @defer.inlineCallbacks
    def test_sourceStamp_prop(self):
        ss = {"revision": properties.Property('rev'), "branch": 'dev'}
        yield self.setup_step(trigger.Trigger(schedulerNames=['b'], sourceStamp=ss))
        self.build.setProperty('rev', 602, 'me')
        expected_ss = {"revision": 602, "branch": 'dev'}
        self.expect_outcome(result=SUCCESS)
        self.expectTriggeredWith(b=(False, [expected_ss], {}))
        yield self.run_step()

    @defer.inlineCallbacks
    def test_waitForFinish(self):
        yield self.setup_step(trigger.Trigger(schedulerNames=['a', 'b'], waitForFinish=True))
        self.expect_outcome(result=SUCCESS, state_string='triggered a, b')
        self.expectTriggeredWith(a=(True, [], {}), b=(True, [], {}))
        self.expectTriggeredLinks('a', 'b')
        yield self.run_step()

    @defer.inlineCallbacks
    def test_waitForFinish_failure(self):
        yield self.setup_step(trigger.Trigger(schedulerNames=['a'], waitForFinish=True))
        self.scheduler_a.result = FAILURE
        self.expect_outcome(result=FAILURE)
        self.expectTriggeredWith(a=(True, [], {}))
        self.expectTriggeredLinks('afailed')
        yield self.run_step(results_dict={11: FAILURE})

    @defer.inlineCallbacks
    def test_waitForFinish_split_failure(self):
        yield self.setup_step(trigger.Trigger(schedulerNames=['a', 'b'], waitForFinish=True))
        self.scheduler_a.result = FAILURE
        self.scheduler_b.result = SUCCESS
        self.expect_outcome(result=FAILURE, state_string='triggered a, b')
        self.expectTriggeredWith(a=(True, [], {}), b=(True, [], {}))
        self.expectTriggeredLinks('afailed', 'b')
        yield self.run_step(results_dict={11: FAILURE})

    @defer.inlineCallbacks
    def test_waitForFinish_exception(self):
        yield self.setup_step(trigger.Trigger(schedulerNames=['a', 'b'], waitForFinish=True))
        self.get_nth_step(0).addCompleteLog = Mock()
        self.scheduler_b.exception = True
        self.expect_outcome(result=EXCEPTION, state_string='triggered a, b')
        self.expectTriggeredWith(a=(True, [], {}), b=(True, [], {}))
        self.expectTriggeredLinks('a')  # b doesn't return a brid
        yield self.run_step()
        self.assertEqual(len(self.get_nth_step(0).addCompleteLog.call_args_list), 1)

    @defer.inlineCallbacks
    def test_virtual_builder(self):
        yield self.setup_step(trigger.Trigger(schedulerNames=['c'], waitForFinish=True))
        self.expect_outcome(result=SUCCESS, state_string='triggered c')
        self.expectTriggeredWith(c=(True, [], {}))
        self.expectTriggeredLinks('c')
        yield self.run_step()

    @defer.inlineCallbacks
    def test_set_properties(self):
        yield self.setup_step(
            trigger.Trigger(schedulerNames=['a'], set_properties={"x": 1, "y": 2})
        )
        self.expect_outcome(result=SUCCESS)
        self.expectTriggeredWith(a=(False, [], {"x": (1, 'Trigger'), "y": (2, 'Trigger')}))
        yield self.run_step()

    @defer.inlineCallbacks
    def test_set_properties_prop(self):
        yield self.setup_step(
            trigger.Trigger(
                schedulerNames=['a'], set_properties={"x": properties.Property('X'), "y": 2}
            )
        )
        self.build.setProperty('X', 'xxx', 'here')
        self.expect_outcome(result=SUCCESS)
        self.expectTriggeredWith(a=(False, [], {"x": ('xxx', 'Trigger'), "y": (2, 'Trigger')}))
        yield self.run_step()

    @defer.inlineCallbacks
    def test_copy_properties(self):
        yield self.setup_step(trigger.Trigger(schedulerNames=['a'], copy_properties=['a', 'b']))
        self.build.setProperty('a', 'A', 'AA')
        self.build.setProperty('b', 'B', 'BB')
        self.build.setProperty('c', 'C', 'CC')
        self.expect_outcome(result=SUCCESS)
        self.expectTriggeredWith(a=(False, [], {"a": ('A', 'Trigger'), "b": ('B', 'Trigger')}))
        yield self.run_step()

    @defer.inlineCallbacks
    def test_waitForFinish_interrupt(self):
        yield self.setup_step(trigger.Trigger(schedulerNames=['a'], waitForFinish=True))

        self.expect_outcome(result=CANCELLED, state_string='interrupted')
        self.expectTriggeredWith(a=(True, [], {}))
        d = self.run_step()

        # interrupt before the callLater representing the Triggerable
        # schedulers completes
        self.get_nth_step(0).interrupt(failure.Failure(RuntimeError('oh noes')))

        yield d

    @defer.inlineCallbacks
    def test_waitForFinish_interrupt_no_connection(self):
        yield self.setup_step(trigger.Trigger(schedulerNames=['a'], waitForFinish=True))

        self.expect_outcome(result=CANCELLED, state_string='interrupted')
        self.expectTriggeredWith(a=(True, [], {}))
        self.scheduler_a.never_finish = True
        d = self.run_step()

        # interrupt before the callLater representing the Triggerable
        # schedulers completes
        self.build.conn = None
        self.get_nth_step(0).interrupt(failure.Failure(RuntimeError('oh noes')))

        yield d

    @defer.inlineCallbacks
    def test_getSchedulersAndProperties_back_comp(self):
        class DynamicTrigger(trigger.Trigger):
            def getSchedulersAndProperties(self):
                return [("a", {}, False), ("b", {}, True)]

        yield self.setup_step(DynamicTrigger(schedulerNames=['a', 'b']))
        self.scheduler_a.result = SUCCESS
        self.scheduler_b.result = FAILURE
        self.expect_outcome(result=SUCCESS, state_string='triggered a, b')
        self.expectTriggeredWith(a=(False, [], {}), b=(False, [], {}))
        yield self.run_step()

    @defer.inlineCallbacks
    def test_unimportantSchedulerNames(self):
        yield self.setup_step(
            trigger.Trigger(schedulerNames=['a', 'b'], unimportantSchedulerNames=['b'])
        )
        self.scheduler_a.result = SUCCESS
        self.scheduler_b.result = FAILURE
        self.expect_outcome(result=SUCCESS, state_string='triggered a, b')
        self.expectTriggeredWith(a=(False, [], {}), b=(False, [], {}))
        yield self.run_step()

    @defer.inlineCallbacks
    def test_unimportantSchedulerNames_with_more_brids_for_bsid(self):
        yield self.setup_step(
            trigger.Trigger(schedulerNames=['a', 'c'], unimportantSchedulerNames=['c'])
        )
        self.scheduler_a.result = SUCCESS
        self.scheduler_c.result = FAILURE
        self.expect_outcome(result=SUCCESS, state_string='triggered a, c')
        self.expectTriggeredWith(a=(False, [], {}), c=(False, [], {}))
        yield self.run_step()



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

import sqlalchemy as sa
from twisted.trial import unittest

from buildbot.db.types.json import JsonObject
from buildbot.test.util import migration
from buildbot.util import sautils


class Migration(migration.MigrateTestMixin, unittest.TestCase):
    def setUp(self):
        return self.setUpMigrateTest()

    def create_tables_thd(self, conn):
        metadata = sa.MetaData()
        metadata.bind = conn

        workers = sautils.Table(
            "workers",
            metadata,
            sa.Column("id", sa.Integer, primary_key=True),
            sa.Column("name", sa.String(50), nullable=False),
            sa.Column("info", JsonObject, nullable=False),
            sa.Column("paused", sa.SmallInteger, nullable=False, server_default="0"),
            sa.Column("graceful", sa.SmallInteger, nullable=False, server_default="0"),
        )
        workers.create(bind=conn)

        conn.execute(
            workers.insert(),
            [
                {
                    "id": 4,
                    "name": "worker1",
                    "info": "{\"key\": \"value\"}",
                    "paused": 0,
                    "graceful": 0,
                }
            ],
        )
        conn.commit()

    def test_update(self):
        def setup_thd(conn):
            self.create_tables_thd(conn)

        def verify_thd(conn):
            metadata = sa.MetaData()
            metadata.bind = conn

            workers = sautils.Table('workers', metadata, autoload_with=conn)
            self.assertIsInstance(workers.c.pause_reason.type, sa.Text)

            q = sa.select(
                workers.c.name,
                workers.c.pause_reason,
            )

            num_rows = 0
            for row in conn.execute(q):
                self.assertEqual(row.name, "worker1")
                self.assertIsNone(row.pause_reason)
                num_rows += 1
            self.assertEqual(num_rows, 1)

        return self.do_test_migration('063', '064', setup_thd, verify_thd)



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members
# Copyright Manba Team

from twisted.internet import defer
from twisted.trial import unittest

from buildbot.test.fake.web import FakeRequest
from buildbot.test.fake.web import fakeMasterForHooks
from buildbot.test.reactor import TestReactorMixin
from buildbot.www import change_hook
from buildbot.www.hooks.bitbucket import _HEADER_EVENT

gitJsonPayload = b"""{
    "canon_url": "https://bitbucket.org",
    "commits": [
        {
            "author": "marcus",
            "branch": "master",
            "files": [
                {
                    "file": "somefile.py",
                    "type": "modified"
                }
            ],
            "message": "Added some more things to somefile.py",
            "node": "620ade18607a",
            "parents": [
                "702c70160afc"
            ],
            "raw_author": "Marcus Bertrand <marcus@somedomain.com>",
            "raw_node": "620ade18607ac42d872b568bb92acaa9a28620e9",
            "revision": null,
            "size": -1,
            "timestamp": "2012-05-30 05:58:56",
            "utctimestamp": "2012-05-30 03:58:56+00:00"
        }
    ],
    "repository": {
        "absolute_url": "/marcus/project-x/",
        "fork": false,
        "is_private": true,
        "name": "Project X",
        "owner": "marcus",
        "scm": "git",
        "slug": "project-x",
        "website": "https://atlassian.com/"
    },
    "user": "marcus"
}"""

mercurialJsonPayload = b"""{
    "canon_url": "https://bitbucket.org",
    "commits": [
        {
            "author": "marcus",
            "branch": "master",
            "files": [
                {
                    "file": "somefile.py",
                    "type": "modified"
                }
            ],
            "message": "Added some more things to somefile.py",
            "node": "620ade18607a",
            "parents": [
                "702c70160afc"
            ],
            "raw_author": "Marcus Bertrand <marcus@somedomain.com>",
            "raw_node": "620ade18607ac42d872b568bb92acaa9a28620e9",
            "revision": null,
            "size": -1,
            "timestamp": "2012-05-30 05:58:56",
            "utctimestamp": "2012-05-30 03:58:56+00:00"
        }
    ],
    "repository": {
        "absolute_url": "/marcus/project-x/",
        "fork": false,
        "is_private": true,
        "name": "Project X",
        "owner": "marcus",
        "scm": "hg",
        "slug": "project-x",
        "website": "https://atlassian.com/"
    },
    "user": "marcus"
}"""

gitJsonNoCommitsPayload = b"""{
    "canon_url": "https://bitbucket.org",
    "commits": [
    ],
    "repository": {
        "absolute_url": "/marcus/project-x/",
        "fork": false,
        "is_private": true,
        "name": "Project X",
        "owner": "marcus",
        "scm": "git",
        "slug": "project-x",
        "website": "https://atlassian.com/"
    },
    "user": "marcus"
}"""

mercurialJsonNoCommitsPayload = b"""{
    "canon_url": "https://bitbucket.org",
    "commits": [
    ],
    "repository": {
        "absolute_url": "/marcus/project-x/",
        "fork": false,
        "is_private": true,
        "name": "Project X",
        "owner": "marcus",
        "scm": "hg",
        "slug": "project-x",
        "website": "https://atlassian.com/"
    },
    "user": "marcus"
}"""


class TestChangeHookConfiguredWithBitbucketChange(TestReactorMixin, unittest.TestCase):
    """Unit tests for BitBucket Change Hook"""

    @defer.inlineCallbacks
    def setUp(self):
        self.setup_test_reactor()
        master = yield fakeMasterForHooks(self)
        self.change_hook = change_hook.ChangeHookResource(
            dialects={'bitbucket': True}, master=master
        )

    @defer.inlineCallbacks
    def testGitWithChange(self):
        change_dict = {b'payload': [gitJsonPayload]}

        request = FakeRequest(change_dict)
        request.received_headers[_HEADER_EVENT] = b"repo:push"
        request.uri = b'/change_hook/bitbucket'
        request.method = b'POST'

        yield request.test_render(self.change_hook)

        self.assertEqual(len(self.change_hook.master.data.updates.changesAdded), 1)
        commit = self.change_hook.master.data.updates.changesAdded[0]

        self.assertEqual(commit['files'], ['somefile.py'])
        self.assertEqual(commit['repository'], 'https://bitbucket.org/marcus/project-x/')
        self.assertEqual(commit['when_timestamp'], 1338350336)
        self.assertEqual(commit['author'], 'Marcus Bertrand <marcus@somedomain.com>')
        self.assertEqual(commit['revision'], '620ade18607ac42d872b568bb92acaa9a28620e9')
        self.assertEqual(commit['comments'], 'Added some more things to somefile.py')
        self.assertEqual(commit['branch'], 'master')
        self.assertEqual(
            commit['revlink'],
            'https://bitbucket.org/marcus/project-x/commits/'
            '620ade18607ac42d872b568bb92acaa9a28620e9',
        )
        self.assertEqual(commit['properties']['event'], 'repo:push')

    @defer.inlineCallbacks
    def testGitWithNoCommitsPayload(self):
        change_dict = {b'payload': [gitJsonNoCommitsPayload]}

        request = FakeRequest(change_dict)
        request.uri = b'/change_hook/bitbucket'
        request.method = b'POST'

        yield request.test_render(self.change_hook)

        self.assertEqual(len(self.change_hook.master.data.updates.changesAdded), 0)
        self.assertEqual(request.written, b'no change found')

    @defer.inlineCallbacks
    def testMercurialWithChange(self):
        change_dict = {b'payload': [mercurialJsonPayload]}

        request = FakeRequest(change_dict)
        request.received_headers[_HEADER_EVENT] = b"repo:push"
        request.uri = b'/change_hook/bitbucket'
        request.method = b'POST'

        yield request.test_render(self.change_hook)

        self.assertEqual(len(self.change_hook.master.data.updates.changesAdded), 1)
        commit = self.change_hook.master.data.updates.changesAdded[0]

        self.assertEqual(commit['files'], ['somefile.py'])
        self.assertEqual(commit['repository'], 'https://bitbucket.org/marcus/project-x/')
        self.assertEqual(commit['when_timestamp'], 1338350336)
        self.assertEqual(commit['author'], 'Marcus Bertrand <marcus@somedomain.com>')
        self.assertEqual(commit['revision'], '620ade18607ac42d872b568bb92acaa9a28620e9')
        self.assertEqual(commit['comments'], 'Added some more things to somefile.py')
        self.assertEqual(commit['branch'], 'master')
        self.assertEqual(
            commit['revlink'],
            'https://bitbucket.org/marcus/project-x/commits/'
            '620ade18607ac42d872b568bb92acaa9a28620e9',
        )
        self.assertEqual(commit['properties']['event'], 'repo:push')

    @defer.inlineCallbacks
    def testMercurialWithNoCommitsPayload(self):
        change_dict = {b'payload': [mercurialJsonNoCommitsPayload]}

        request = FakeRequest(change_dict)
        request.uri = b'/change_hook/bitbucket'
        request.method = b'POST'

        yield request.test_render(self.change_hook)

        self.assertEqual(len(self.change_hook.master.data.updates.changesAdded), 0)
        self.assertEqual(request.written, b'no change found')

    @defer.inlineCallbacks
    def testWithNoJson(self):
        request = FakeRequest()
        request.uri = b'/change_hook/bitbucket'
        request.method = b'POST'

        yield request.test_render(self.change_hook)
        self.assertEqual(len(self.change_hook.master.data.updates.changesAdded), 0)
        self.assertEqual(request.written, b'Error processing changes.')
        request.setResponseCode.assert_called_with(500, b'Error processing changes.')
        self.assertEqual(len(self.flushLoggedErrors()), 1)

    @defer.inlineCallbacks
    def testGitWithChangeAndProject(self):
        change_dict = {b'payload': [gitJsonPayload], b'project': [b'project-name']}

        request = FakeRequest(change_dict)
        request.uri = b'/change_hook/bitbucket'
        request.method = b'POST'

        yield request.test_render(self.change_hook)

        self.assertEqual(len(self.change_hook.master.data.updates.changesAdded), 1)
        commit = self.change_hook.master.data.updates.changesAdded[0]

        self.assertEqual(commit['project'], 'project-name')



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

import datetime
import textwrap

from dateutil.tz import tzutc
from parameterized import parameterized
from twisted.internet import defer
from twisted.trial import unittest

from buildbot.process.results import FAILURE
from buildbot.process.results import RETRY
from buildbot.process.results import SUCCESS
from buildbot.reporters import utils
from buildbot.test import fakedb
from buildbot.test.fake import fakemaster
from buildbot.test.reactor import TestReactorMixin
from buildbot.test.util import logging


def sort_builds(builds):
    return sorted(builds, key=lambda key: key['buildid'])


class TestDataUtils(TestReactorMixin, unittest.TestCase, logging.LoggingMixin):
    LOGCONTENT = textwrap.dedent("""\
        line zero
        line 1""")

    @defer.inlineCallbacks
    def setUp(self):
        self.setup_test_reactor()
        self.master = yield fakemaster.make_master(self, wantData=True, wantDb=True, wantMq=True)

    @defer.inlineCallbacks
    def setupDb(self):
        yield self.master.db.insert_test_data([
            fakedb.Master(id=92),
            fakedb.Worker(id=13, name='wrk'),
            fakedb.Buildset(id=97, results=SUCCESS, reason="testReason0"),
            fakedb.Buildset(id=98, results=SUCCESS, reason="testReason1"),
            fakedb.Buildset(id=99, results=SUCCESS, reason="testReason2", parent_buildid=21),
            fakedb.Builder(id=80, name='Builder1'),
            fakedb.Builder(id=81, name='Builder2'),
            fakedb.BuildRequest(id=9, buildsetid=97, builderid=80),
            fakedb.BuildRequest(id=10, buildsetid=97, builderid=80),
            fakedb.BuildRequest(id=11, buildsetid=98, builderid=80),
            fakedb.BuildRequest(id=12, buildsetid=98, builderid=80),
            fakedb.BuildRequest(id=13, buildsetid=99, builderid=81),
            fakedb.Build(
                id=18,
                number=0,
                builderid=80,
                buildrequestid=9,
                workerid=13,
                masterid=92,
                results=FAILURE,
            ),
            fakedb.Build(
                id=19,
                number=1,
                builderid=80,
                buildrequestid=10,
                workerid=13,
                masterid=92,
                results=RETRY,
            ),
            fakedb.Build(
                id=20,
                number=2,
                builderid=80,
                buildrequestid=11,
                workerid=13,
                masterid=92,
                results=SUCCESS,
            ),
            fakedb.Build(
                id=21,
                number=3,
                builderid=80,
                buildrequestid=12,
                workerid=13,
                masterid=92,
                results=SUCCESS,
            ),
            fakedb.Build(
                id=22,
                number=1,
                builderid=81,
                buildrequestid=13,
                workerid=13,
                masterid=92,
                results=SUCCESS,
            ),
            fakedb.BuildsetSourceStamp(buildsetid=98, sourcestampid=234),
            fakedb.SourceStamp(id=234),
            fakedb.Change(
                changeid=13,
                branch='trunk',
                revision='9283',
                author='me@foo',
                repository='svn://...',
                codebase='cbsvn',
                project='world-domination',
                sourcestampid=234,
            ),
            fakedb.Patch(
                id=99,
                patch_base64='aGVsbG8sIHdvcmxk',
                patch_author='him@foo',
                patch_comment='foo',
                subdir='/foo',
                patchlevel=3,
            ),
            fakedb.SourceStamp(id=235, patchid=99),
        ])
        for _id in (20, 21):
            yield self.master.db.insert_test_data([
                fakedb.BuildProperty(buildid=_id, name="workername", value="wrk"),
                fakedb.BuildProperty(buildid=_id, name="reason", value="because"),
                fakedb.BuildProperty(buildid=_id, name="owner", value="him"),
                fakedb.Step(id=100 + _id, number=29, buildid=_id, name="step1"),
                fakedb.Step(id=200 + _id, number=30, buildid=_id, name="step2"),
                fakedb.Log(
                    id=60 + _id, stepid=100 + _id, name='stdio', slug='stdio', type='s', num_lines=2
                ),
                fakedb.LogChunk(
                    logid=60 + _id, first_line=0, last_line=1, compressed=0, content=self.LOGCONTENT
                ),
            ])

        @defer.inlineCallbacks
        def getChangesForBuild(buildid):
            assert buildid == 20
            ch = yield self.master.db.changes.getChange(13)
            if ch is None:
                return []
            return [ch]

        self.master.db.changes.getChangesForBuild = getChangesForBuild

    @parameterized.expand([
        ('disabled', 'log', 'step', False, False),
        ('enabled', 'log', 'step', True, True),
        ('empty_list', 'log', 'step', [], False),
        ('name_no_match', 'log', 'step', ['log2'], False),
        ('name_match', 'log', 'step', ['log'], True),
        ('name_match_step_no_match', 'log', 'step', ['step2.log'], False),
        ('name_no_match_step_match', 'log', 'step', ['step.log2'], False),
        ('name_match_step_match', 'log', 'step', ['step.log'], True),
    ])
    def test_should_attach_log(self, name, log_name, step_name, log_config, expected):
        log = {'name': log_name, 'stepname': step_name}
        self.assertEqual(utils.should_attach_log(log_config, log), expected)

    @defer.inlineCallbacks
    def test_getDetailsForBuildset(self):
        yield self.setupDb()
        res = yield utils.getDetailsForBuildset(
            self.master, 98, want_properties=True, want_steps=True, want_previous_build=True
        )
        self.assertEqual(len(res['builds']), 2)

        res['builds'] = sort_builds(res['builds'])

        build1 = res['builds'][0]
        build2 = res['builds'][1]
        buildset = res['buildset']
        self.assertEqual(
            build1['properties'],
            {
                'reason': ('because', 'fakedb'),
                'owner': ('him', 'fakedb'),
                'workername': ('wrk', 'fakedb'),
            },
        )
        self.assertEqual(len(build1['steps']), 2)
        self.assertEqual(build1['buildid'], 20)
        self.assertEqual(build2['buildid'], 21)
        self.assertEqual(buildset['bsid'], 98)

        # make sure prev_build was computed
        self.assertEqual(build1['prev_build']['buildid'], 18)
        self.assertEqual(build2['prev_build']['buildid'], 20)

    @defer.inlineCallbacks
    def test_getDetailsForBuild(self):
        yield self.setupDb()
        build = yield self.master.data.get(("builds", 21))
        yield utils.getDetailsForBuild(
            self.master,
            build,
            want_properties=False,
            want_steps=False,
            want_previous_build=False,
            want_logs=False,
        )

        self.assertEqual(build['parentbuild'], None)
        self.assertEqual(build['parentbuilder'], None)

    @defer.inlineCallbacks
    def test_getDetailsForBuildWithParent(self):
        yield self.setupDb()
        build = yield self.master.data.get(("builds", 22))
        yield utils.getDetailsForBuild(
            self.master,
            build,
            want_properties=False,
            want_steps=False,
            want_previous_build=False,
            want_logs=False,
        )

        self.assertEqual(build['parentbuild']['buildid'], 21)
        self.assertEqual(build['parentbuilder']['name'], "Builder1")

    @defer.inlineCallbacks
    def test_getDetailsForBuildsetWithLogs(self):
        yield self.setupDb()
        res = yield utils.getDetailsForBuildset(
            self.master,
            98,
            want_properties=True,
            want_steps=True,
            want_previous_build=True,
            want_logs=True,
            want_logs_content=True,
        )

        build1 = sort_builds(res['builds'])[0]
        self.assertEqual(
            build1['steps'][0]['logs'][0]['content']['content'], self.LOGCONTENT + "\n"
        )
        self.assertEqual(
            build1['steps'][0]['logs'][0]['url'],
            'http://localhost:8080/#/builders/80/builds/2/steps/29/logs/stdio',
        )

    @defer.inlineCallbacks
    def test_get_details_for_buildset_all(self):
        yield self.setupDb()
        self.maxDiff = None
        res = yield utils.getDetailsForBuildset(
            self.master,
            98,
            want_properties=True,
            want_steps=True,
            want_previous_build=True,
            want_logs=True,
            want_logs_content=True,
        )

        res['builds'] = sort_builds(res['builds'])

        self.assertEqual(
            res,
            {
                'builds': [
                    {
                        'builder': {
                            'builderid': 80,
                            'description': None,
                            'description_format': None,
                            'description_html': None,
                            'projectid': None,
                            'masterids': [],
                            'name': 'Builder1',
                            'tags': [],
                        },
                        'builderid': 80,
                        'buildid': 20,
                        'buildrequestid': 11,
                        'buildset': {
                            'bsid': 98,
                            'complete': False,
                            'complete_at': None,
                            'external_idstring': 'extid',
                            'rebuilt_buildid': None,
                            'parent_buildid': None,
                            'parent_relationship': None,
                            'reason': 'testReason1',
                            'results': 0,
                            'sourcestamps': [
                                {
                                    'branch': 'master',
                                    'codebase': '',
                                    'created_at': datetime.datetime(
                                        1972, 11, 5, 18, 7, 14, tzinfo=tzutc()
                                    ),
                                    'patch': None,
                                    'project': 'proj',
                                    'repository': 'repo',
                                    'revision': 'rev-234',
                                    'ssid': 234,
                                }
                            ],
                            'submitted_at': 12345678,
                        },
                        'complete': False,
                        'complete_at': None,
                        "locks_duration_s": 0,
                        'masterid': 92,
                        'number': 2,
                        'prev_build': {
                            'builderid': 80,
                            'buildid': 18,
                            'buildrequestid': 9,
                            'complete': False,
                            'complete_at': None,
                            "locks_duration_s": 0,
                            'masterid': 92,
                            'number': 0,
                            'properties': {},
                            'results': 2,
                            'started_at': datetime.datetime(2011, 5, 1, 15, 3, 42, tzinfo=tzutc()),
                            'state_string': 'test',
                            'workerid': 13,
                        },
                        'properties': {
                            'owner': ('him', 'fakedb'),
                            'reason': ('because', 'fakedb'),
                            'workername': ('wrk', 'fakedb'),
                        },
                        'results': 0,
                        'started_at': datetime.datetime(2011, 5, 1, 15, 3, 42, tzinfo=tzutc()),
                        'state_string': 'test',
                        'steps': [
                            {
                                'buildid': 20,
                                'complete': False,
                                'complete_at': None,
                                'hidden': False,
                                "locks_acquired_at": None,
                                'logs': [
                                    {
                                        'complete': False,
                                        'content': {
                                            'content': 'line zero\nline 1\n',
                                            'firstline': 0,
                                            'logid': 80,
                                        },
                                        'logid': 80,
                                        'name': 'stdio',
                                        'num_lines': 2,
                                        'slug': 'stdio',
                                        'stepid': 120,
                                        'stepname': 'step1',
                                        'type': 's',
                                        'url': 'http://localhost:8080/#/builders/80/builds/2/steps/29/logs/stdio',
                                        'url_raw': 'http://localhost:8080/api/v2/logs/80/raw',
                                        'url_raw_inline': 'http://localhost:8080/api/v2/logs/80/raw_inline',
                                    }
                                ],
                                'name': 'step1',
                                'number': 29,
                                'results': None,
                                'started_at': datetime.datetime(
                                    2011, 5, 1, 15, 3, 42, tzinfo=tzutc()
                                ),
                                'state_string': '',
                                'stepid': 120,
                                'urls': [],
                            },
                            {
                                'buildid': 20,
                                'complete': False,
                                'complete_at': None,
                                'hidden': False,
                                "locks_acquired_at": None,
                                'logs': [],
                                'name': 'step2',
                                'number': 30,
                                'results': None,
                                'started_at': datetime.datetime(
                                    2011, 5, 1, 15, 3, 42, tzinfo=tzutc()
                                ),
                                'state_string': '',
                                'stepid': 220,
                                'urls': [],
                            },
                        ],
                        'url': 'http://localhost:8080/#/builders/80/builds/2',
                        'workerid': 13,
                    },
                    {
                        'builder': {
                            'builderid': 80,
                            'description': None,
                            'description_format': None,
                            'description_html': None,
                            'projectid': None,
                            'masterids': [],
                            'name': 'Builder1',
                            'tags': [],
                        },
                        'builderid': 80,
                        'buildid': 21,
                        'buildrequestid': 12,
                        'buildset': {
                            'bsid': 98,
                            'complete': False,
                            'complete_at': None,
                            'external_idstring': 'extid',
                            'rebuilt_buildid': None,
                            'parent_buildid': None,
                            'parent_relationship': None,
                            'reason': 'testReason1',
                            'results': 0,
                            'sourcestamps': [
                                {
                                    'branch': 'master',
                                    'codebase': '',
                                    'created_at': datetime.datetime(
                                        1972, 11, 5, 18, 7, 14, tzinfo=tzutc()
                                    ),
                                    'patch': None,
                                    'project': 'proj',
                                    'repository': 'repo',
                                    'revision': 'rev-234',
                                    'ssid': 234,
                                }
                            ],
                            'submitted_at': 12345678,
                        },
                        'complete': False,
                        'complete_at': None,
                        "locks_duration_s": 0,
                        'masterid': 92,
                        'number': 3,
                        'prev_build': {
                            'builderid': 80,
                            'buildid': 20,
                            'buildrequestid': 11,
                            'complete': False,
                            'complete_at': None,
                            "locks_duration_s": 0,
                            'masterid': 92,
                            'number': 2,
                            'properties': {},
                            'results': 0,
                            'started_at': datetime.datetime(2011, 5, 1, 15, 3, 42, tzinfo=tzutc()),
                            'state_string': 'test',
                            'workerid': 13,
                        },
                        'properties': {
                            'owner': ('him', 'fakedb'),
                            'reason': ('because', 'fakedb'),
                            'workername': ('wrk', 'fakedb'),
                        },
                        'results': 0,
                        'started_at': datetime.datetime(2011, 5, 1, 15, 3, 42, tzinfo=tzutc()),
                        'state_string': 'test',
                        'steps': [
                            {
                                'buildid': 21,
                                'complete': False,
                                'complete_at': None,
                                'hidden': False,
                                "locks_acquired_at": None,
                                'logs': [
                                    {
                                        'complete': False,
                                        'content': {
                                            'content': 'line zero\nline 1\n',
                                            'firstline': 0,
                                            'logid': 81,
                                        },
                                        'logid': 81,
                                        'name': 'stdio',
                                        'num_lines': 2,
                                        'slug': 'stdio',
                                        'stepid': 121,
                                        'stepname': 'step1',
                                        'type': 's',
                                        'url': 'http://localhost:8080/#/builders/80/builds/3/steps/29/logs/stdio',
                                        'url_raw': 'http://localhost:8080/api/v2/logs/81/raw',
                                        'url_raw_inline': 'http://localhost:8080/api/v2/logs/81/raw_inline',
                                    }
                                ],
                                'name': 'step1',
                                'number': 29,
                                'results': None,
                                'started_at': datetime.datetime(
                                    2011, 5, 1, 15, 3, 42, tzinfo=tzutc()
                                ),
                                'state_string': '',
                                'stepid': 121,
                                'urls': [],
                            },
                            {
                                'buildid': 21,
                                'complete': False,
                                'complete_at': None,
                                'hidden': False,
                                "locks_acquired_at": None,
                                'logs': [],
                                'name': 'step2',
                                'number': 30,
                                'results': None,
                                'started_at': datetime.datetime(
                                    2011, 5, 1, 15, 3, 42, tzinfo=tzutc()
                                ),
                                'state_string': '',
                                'stepid': 221,
                                'urls': [],
                            },
                        ],
                        'url': 'http://localhost:8080/#/builders/80/builds/3',
                        'workerid': 13,
                    },
                ],
                'buildset': {
                    'bsid': 98,
                    'complete': False,
                    'complete_at': None,
                    'external_idstring': 'extid',
                    'rebuilt_buildid': None,
                    'parent_buildid': None,
                    'parent_relationship': None,
                    'reason': 'testReason1',
                    'results': 0,
                    'sourcestamps': [
                        {
                            'branch': 'master',
                            'codebase': '',
                            'created_at': datetime.datetime(1972, 11, 5, 18, 7, 14, tzinfo=tzutc()),
                            'patch': None,
                            'project': 'proj',
                            'repository': 'repo',
                            'revision': 'rev-234',
                            'ssid': 234,
                        }
                    ],
                    'submitted_at': 12345678,
                },
            },
        )

    @defer.inlineCallbacks
    def test_getResponsibleUsers(self):
        yield self.setupDb()
        res = yield utils.getResponsibleUsersForSourceStamp(self.master, 234)
        self.assertEqual(res, ["me@foo"])

    @defer.inlineCallbacks
    def test_getResponsibleUsersFromPatch(self):
        yield self.setupDb()
        res = yield utils.getResponsibleUsersForSourceStamp(self.master, 235)
        self.assertEqual(res, ["him@foo"])

    @defer.inlineCallbacks
    def test_getResponsibleUsersForBuild(self):
        yield self.setupDb()
        res = yield utils.getResponsibleUsersForBuild(self.master, 20)
        self.assertEqual(sorted(res), sorted(["me@foo", "him"]))

    @defer.inlineCallbacks
    def test_getResponsibleUsersForBuildWithBadOwner(self):
        self.setUpLogging()
        yield self.setupDb()
        yield self.master.db.insert_test_data([
            fakedb.BuildProperty(buildid=20, name="owner", value=["him"]),
        ])
        res = yield utils.getResponsibleUsersForBuild(self.master, 20)
        self.assertLogged("Please report a bug")
        self.assertEqual(sorted(res), sorted(["me@foo", "him"]))

    @defer.inlineCallbacks
    def test_getResponsibleUsersForBuildWithOwners(self):
        yield self.setupDb()
        yield self.master.db.insert_test_data([
            fakedb.BuildProperty(buildid=20, name="owners", value=["him", "her"]),
        ])
        res = yield utils.getResponsibleUsersForBuild(self.master, 20)
        self.assertEqual(sorted(res), sorted(["me@foo", "him", "her"]))

    @defer.inlineCallbacks
    def test_get_responsible_users_for_buildset_with_owner(self):
        yield self.setupDb()

        yield self.master.db.insert_test_data([
            fakedb.BuildsetProperty(
                buildsetid=98, property_name="owner", property_value='["buildset_owner", "fakedb"]'
            ),
        ])

        res = yield utils.get_responsible_users_for_buildset(self.master, 98)
        self.assertEqual(sorted(res), sorted(["buildset_owner"]))

    @defer.inlineCallbacks
    def test_get_responsible_users_for_buildset_no_owner(self):
        yield self.setupDb()
        res = yield utils.get_responsible_users_for_buildset(self.master, 99)
        self.assertEqual(sorted(res), sorted([]))

    @defer.inlineCallbacks
    def test_getPreviousBuild(self):
        yield self.setupDb()
        build = yield self.master.data.get(("builds", 21))
        res = yield utils.getPreviousBuild(self.master, build)
        self.assertEqual(res['buildid'], 20)

    @defer.inlineCallbacks
    def test_getPreviousBuildWithRetry(self):
        yield self.setupDb()
        build = yield self.master.data.get(("builds", 20))
        res = yield utils.getPreviousBuild(self.master, build)
        self.assertEqual(res['buildid'], 18)


class TestURLUtils(TestReactorMixin, unittest.TestCase):
    @defer.inlineCallbacks
    def setUp(self):
        self.setup_test_reactor()
        self.master = yield fakemaster.make_master(self)

    def test_UrlForBuild(self):
        self.assertEqual(
            utils.getURLForBuild(self.master, 1, 3), 'http://localhost:8080/#/builders/1/builds/3'
        )



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members
from __future__ import annotations

import enum

from twisted.internet import defer
from twisted.python.filepath import FilePath
from twisted.trial.unittest import SkipTest

from buildbot.test.fake.worker import SeverWorkerConnectionMixin
from buildbot.test.fake.worker import disconnect_master_side_worker
from buildbot.util.twisted import async_to_deferred
from buildbot.worker import AbstractLatentWorker

RemoteWorker: type | None = None
try:
    from buildbot_worker.bot import LocalWorker as RemoteWorker
    from buildbot_worker.pb import BotPbLike
except ImportError:
    pass


class States(enum.Enum):
    STOPPED = 0
    STARTING = 1
    STARTED = 2
    STOPPING = 3


class LatentController(SeverWorkerConnectionMixin):
    """
    A controller for ``ControllableLatentWorker``.

    https://glyph.twistedmatrix.com/2015/05/separate-your-fakes-and-your-inspectors.html

    Note that by default workers will connect automatically if True is passed
    to start_instance().

    Also by default workers will disconnect automatically just as
    stop_instance() is executed.
    """

    def __init__(
        self,
        case,
        name,
        kind=None,
        build_wait_timeout=600,
        starts_without_substantiate=None,
        **kwargs,
    ):
        self.case = case
        self.build_wait_timeout = build_wait_timeout
        self.has_crashed = False
        self.worker = ControllableLatentWorker(name, self, **kwargs)
        self.remote_worker = None

        if starts_without_substantiate is not None:
            self.worker.starts_without_substantiate = starts_without_substantiate

        self.state = States.STOPPED
        self.auto_stop_flag = False
        self.auto_start_flag = False
        self.auto_connect_worker = True
        self.auto_disconnect_worker = True

        self._start_deferred = None
        self._stop_deferred = None

        self.kind = kind
        self._started_kind = None
        self._started_kind_deferred = None

    @property
    def starting(self):
        return self.state == States.STARTING

    @property
    def started(self):
        return self.state == States.STARTED

    @property
    def stopping(self):
        return self.state == States.STOPPING

    @property
    def stopped(self):
        return self.state == States.STOPPED

    def auto_start(self, result):
        self.auto_start_flag = result
        if self.auto_start_flag and self.state == States.STARTING:
            self.start_instance(True)

    @async_to_deferred
    async def start_instance(self, result) -> None:
        await self.do_start_instance(result)
        d = self._start_deferred
        self._start_deferred = None
        d.callback(result)

    @async_to_deferred
    async def do_start_instance(self, result) -> None:
        assert self.state == States.STARTING
        self.state = States.STARTED
        if self.auto_connect_worker and result is True:
            await self.connect_worker()

    @async_to_deferred
    async def auto_stop(self, result) -> None:
        self.auto_stop_flag = result
        if self.auto_stop_flag and self.state == States.STOPPING:
            await self.stop_instance(True)

    @async_to_deferred
    async def stop_instance(self, result):
        await self.do_stop_instance()
        d = self._stop_deferred
        self._stop_deferred = None
        d.callback(result)

    @async_to_deferred
    async def do_stop_instance(self):
        assert self.state == States.STOPPING
        self.state = States.STOPPED
        self._started_kind = None
        if self.auto_disconnect_worker:
            await self.disconnect_worker()

    @async_to_deferred
    async def connect_worker(self):
        if self.remote_worker is not None:
            return
        if RemoteWorker is None:
            raise SkipTest("buildbot-worker package is not installed")
        workdir = FilePath(self.case.mktemp())
        workdir.createDirectory()
        self.remote_worker = RemoteWorker(self.worker.name, workdir.path, False)
        await self.remote_worker.setServiceParent(self.worker)

    @async_to_deferred
    async def disconnect_worker(self) -> None:
        await super().disconnect_worker()
        if self.remote_worker is None:
            return

        self.remote_worker, worker = None, self.remote_worker

        disconnect_master_side_worker(self.worker)

        await worker.disownServiceParent()

    def setup_kind(self, build):
        if build:
            self._started_kind_deferred = build.render(self.kind)
        else:
            self._started_kind_deferred = self.kind

    @async_to_deferred
    async def get_started_kind(self):
        if self._started_kind_deferred:
            self._started_kind = await self._started_kind_deferred
            self._started_kind_deferred = None
        return self._started_kind

    def patchBot(self, case, remoteMethod, patch):
        case.patch(BotPbLike, remoteMethod, patch)


class ControllableLatentWorker(AbstractLatentWorker):
    """
    A latent worker that can be controlled by tests.
    """

    builds_may_be_incompatible = True

    def __init__(self, name, controller, **kwargs):
        self._controller = controller
        self._random_password_id = 0
        AbstractLatentWorker.__init__(self, name, None, **kwargs)

    def checkConfig(self, name, _, **kwargs):
        AbstractLatentWorker.checkConfig(
            self, name, None, build_wait_timeout=self._controller.build_wait_timeout, **kwargs
        )

    def reconfigService(self, name, _, **kwargs):
        return super().reconfigService(
            name,
            self.getRandomPass(),
            build_wait_timeout=self._controller.build_wait_timeout,
            **kwargs,
        )

    def _generate_random_password(self):
        self._random_password_id += 1
        return f'password_{self._random_password_id}'

    @async_to_deferred
    async def isCompatibleWithBuild(self, build_props) -> bool:
        if self._controller.state == States.STOPPED:
            return True

        requested_kind = await build_props.render(self._controller.kind)
        curr_kind = await self._controller.get_started_kind()
        return requested_kind == curr_kind

    def start_instance(self, build):
        self._controller.setup_kind(build)

        assert self._controller.state == States.STOPPED
        self._controller.state = States.STARTING

        if self._controller.auto_start_flag:
            self._controller.do_start_instance(True)
            return defer.succeed(True)

        self._controller._start_deferred = defer.Deferred()
        return self._controller._start_deferred

    @async_to_deferred
    async def stop_instance(self, fast: bool = False) -> bool:
        assert self._controller.state == States.STARTED
        self._controller.state = States.STOPPING

        if self._controller.auto_stop_flag:
            await self._controller.do_stop_instance()
            return True
        self._controller._stop_deferred = defer.Deferred()
        return await self._controller._stop_deferred

    def check_instance(self):
        return (not self._controller.has_crashed, "")



#!/usr/bin/python

# this requires python >=2.3 for the 'sets' module.

# The sets.py from python-2.3 appears to work fine under python2.2 . To
# install this script on a host with only python2.2, copy
# /usr/lib/python2.3/sets.py from a newer python into somewhere on your
# PYTHONPATH, then edit the #! line above to invoke python2.2

# python2.1 is right out

# If you run this program as part of your SVN post-commit hooks, it will
# deliver Change notices to a buildmaster that is running a PBChangeSource
# instance.

# edit your svn-repository/hooks/post-commit file, and add lines that look
# like this:

import os
import re
import subprocess
import sys

import sets
from future.utils import text_type
from twisted.cred import credentials
from twisted.internet import defer
from twisted.internet import reactor
from twisted.python import usage
from twisted.spread import pb

'''
# set up PYTHONPATH to contain Twisted/buildbot perhaps, if not already
# installed site-wide
. ~/.environment

/path/to/svn_buildbot.py --repository "$REPOS" --revision "$REV" \
--bbserver localhost --bbport 9989 --username myuser --auth passwd
'''


# We have hackish "-d" handling here rather than in the Options
# subclass below because a common error will be to not have twisted in
# PYTHONPATH; we want to be able to print that error to the log if
# debug mode is on, so we set it up before the imports.

DEBUG = None

if '-d' in sys.argv:
    i = sys.argv.index('-d')
    DEBUG = sys.argv[i + 1]
    del sys.argv[i]
    del sys.argv[i]

if DEBUG:
    f = open(DEBUG, 'a')
    sys.stderr = f
    sys.stdout = f


class Options(usage.Options):
    optParameters = [
        ['repository', 'r', None, "The repository that was changed."],
        ['worker-repo', 'c', None, "In case the repository differs for the workers."],
        ['revision', 'v', None, "The revision that we want to examine (default: latest)"],
        ['bbserver', 's', 'localhost', "The hostname of the server that buildbot is running on"],
        ['bbport', 'p', 8007, "The port that buildbot is listening on"],
        ['username', 'u', 'change', "Username used in PB connection auth"],
        ['auth', 'a', 'changepw', "Password used in PB connection auth"],
        [
            'include',
            'f',
            None,
            '''\
Search the list of changed files for this regular expression, and if there is
at least one match notify buildbot; otherwise buildbot will not do a build.
You may provide more than one -f argument to try multiple
patterns.  If no filter is given, buildbot will always be notified.''',
        ],
        ['filter', 'f', None, "Same as --include.  (Deprecated)"],
        [
            'exclude',
            'F',
            None,
            '''\
The inverse of --filter.  Changed files matching this expression will never
be considered for a build.
You may provide more than one -F argument to try multiple
patterns.  Excludes override includes, that is, patterns that match both an
include and an exclude will be excluded.''',
        ],
        ['encoding', 'e', "utf8", "The encoding of the strings from subversion (default: utf8)"],
        ['project', 'P', None, "The project for the source."],
    ]
    optFlags = [
        ['dryrun', 'n', "Do not actually send changes"],
    ]

    def __init__(self):
        usage.Options.__init__(self)
        self._includes = []
        self._excludes = []
        self['includes'] = None
        self['excludes'] = None

    def opt_include(self, arg):
        self._includes.append(f'.*{arg}.*')

    opt_filter = opt_include

    def opt_exclude(self, arg):
        self._excludes.append(f'.*{arg}.*')

    def postOptions(self):
        if self['repository'] is None:
            raise usage.error("You must pass --repository")
        if self._includes:
            self['includes'] = '({})'.format('|'.join(self._includes))
        if self._excludes:
            self['excludes'] = '({})'.format('|'.join(self._excludes))


def split_file_dummy(changed_file):
    """Split the repository-relative filename into a tuple of (branchname,
    branch_relative_filename). If you have no branches, this should just
    return (None, changed_file).
    """
    return (None, changed_file)


# this version handles repository layouts that look like:
#  trunk/files..                  -> trunk
#  branches/branch1/files..       -> branches/branch1
#  branches/branch2/files..       -> branches/branch2
#


def split_file_branches(changed_file):
    pieces = changed_file.split(os.sep)
    if pieces[0] == 'branches':
        return (os.path.join(*pieces[:2]), os.path.join(*pieces[2:]))
    if pieces[0] == 'trunk':
        return (pieces[0], os.path.join(*pieces[1:]))
    # there are other sibilings of 'trunk' and 'branches'. Pretend they are
    # all just funny-named branches, and let the Schedulers ignore them.
    # return (pieces[0], os.path.join(*pieces[1:]))

    raise RuntimeError(f"cannot determine branch for '{changed_file}'")


split_file = split_file_dummy


class ChangeSender:
    def getChanges(self, opts):
        """Generate and stash a list of Change dictionaries, ready to be sent
        to the buildmaster's PBChangeSource."""

        # first we extract information about the files that were changed
        repo = opts['repository']
        worker_repo = opts['worker-repo'] or repo
        print("Repo:", repo)
        rev_arg = ''
        if opts['revision']:
            rev_arg = '-r {}'.format(opts['revision'])
        changed = subprocess.check_output(f'svnlook changed {rev_arg} "{repo}"', shell=True)
        changed = changed.decode(sys.stdout.encoding)
        changed = changed.split('\n')
        # the first 4 columns can contain status information
        changed = [x[4:] for x in changed]

        message = subprocess.check_output(f'svnlook log {rev_arg} "{repo}"', shell=True)
        message = message.decode(sys.stdout.encoding)
        who = subprocess.check_output(f'svnlook author {rev_arg} "{repo}"', shell=True)
        who = who.decode(sys.stdout.encoding)
        revision = opts.get('revision')
        if revision is not None:
            revision = str(int(revision))

        # see if we even need to notify buildbot by looking at filters first
        changestring = '\n'.join(changed)
        fltpat = opts['includes']
        if fltpat:
            included = sets.Set(re.findall(fltpat, changestring))
        else:
            included = sets.Set(changed)

        expat = opts['excludes']
        if expat:
            excluded = sets.Set(re.findall(expat, changestring))
        else:
            excluded = sets.Set([])
        if len(included.difference(excluded)) == 0:
            print(changestring)
            print(
                f"""\
    Buildbot was not interested, no changes matched any of these filters:\n {fltpat}
    or all the changes matched these exclusions:\n {expat}\
    """
            )
            sys.exit(0)

        # now see which branches are involved
        files_per_branch = {}
        for f in changed:
            branch, filename = split_file(f)
            if branch in files_per_branch.keys():
                files_per_branch[branch].append(filename)
            else:
                files_per_branch[branch] = [filename]

        # now create the Change dictionaries
        changes = []
        encoding = opts['encoding']
        for branch, branch_files in files_per_branch.items():
            d = {
                'who': text_type(who, encoding=encoding),
                'repository': text_type(worker_repo, encoding=encoding),
                'comments': text_type(message, encoding=encoding),
                'revision': revision,
                'project': text_type(opts['project'] or "", encoding=encoding),
                'src': 'svn',
            }
            if branch:
                d['branch'] = text_type(branch, encoding=encoding)
            else:
                d['branch'] = branch

            files = []
            for file in branch_files:
                files.append(text_type(file, encoding=encoding))
            d['files'] = files

            changes.append(d)

        return changes

    def sendChanges(self, opts, changes):
        pbcf = pb.PBClientFactory()
        reactor.connectTCP(opts['bbserver'], int(opts['bbport']), pbcf)
        creds = credentials.UsernamePassword(opts['username'], opts['auth'])
        d = pbcf.login(creds)
        d.addCallback(self.sendAllChanges, changes)
        return d

    def sendAllChanges(self, remote, changes):
        dl = [remote.callRemote('addChange', change) for change in changes]
        return defer.gatherResults(dl, consumeErrors=True)

    def run(self):
        opts = Options()
        try:
            opts.parseOptions()
        except usage.error as ue:
            print(opts)
            print(f"{sys.argv[0]}: {ue}")
            sys.exit()

        changes = self.getChanges(opts)
        if opts['dryrun']:
            for i, c in enumerate(changes):
                print(f"CHANGE #{(i + 1)}")
                keys = sorted(c.keys())
                for k in keys:
                    print(f"[{k:>10}]: {c[k]}")
            print("*NOT* sending any changes")
            return

        d = self.sendChanges(opts, changes)

        def quit(*why):
            print("quitting! because", why)
            reactor.stop()

        d.addCallback(quit, "SUCCESS")

        @d.addErrback
        def failed(f):
            print("FAILURE")
            print(f)
            reactor.stop()

        reactor.callLater(60, quit, "TIMEOUT")
        reactor.run()


if __name__ == '__main__':
    s = ChangeSender()
    s.run()



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from unittest import mock

from twisted.internet import defer
from twisted.trial import unittest

from buildbot.process import cache


class CacheManager(unittest.TestCase):
    def setUp(self):
        self.caches = cache.CacheManager()

    def make_config(self, **kwargs):
        cfg = mock.Mock()
        cfg.caches = kwargs
        return cfg

    def test_get_cache_idempotency(self):
        foo_cache = self.caches.get_cache("foo", None)
        bar_cache = self.caches.get_cache("bar", None)
        foo_cache2 = self.caches.get_cache("foo", None)
        self.assertIdentical(foo_cache, foo_cache2)
        self.assertNotIdentical(foo_cache, bar_cache)

    @defer.inlineCallbacks
    def test_reconfigServiceWithBuildbotConfig(self):
        # load config with one cache loaded and the other not
        foo_cache = self.caches.get_cache("foo", None)
        yield self.caches.reconfigServiceWithBuildbotConfig(self.make_config(foo=5, bar=6, bing=11))

        bar_cache = self.caches.get_cache("bar", None)
        self.assertEqual((foo_cache.max_size, bar_cache.max_size), (5, 6))

    def test_get_metrics(self):
        self.caches.get_cache("foo", None)
        self.assertIn('foo', self.caches.get_metrics())
        metric = self.caches.get_metrics()['foo']
        for k in 'hits', 'refhits', 'misses', 'max_size':
            self.assertIn(k, metric)



# This file is part of Buildbot. Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

import os
import stat
from pathlib import PurePath
from unittest import mock

from parameterized import parameterized
from twisted.internet import defer
from twisted.trial import unittest

from buildbot.process import remotecommand
from buildbot.test.fake import fakemaster
from buildbot.test.reactor import TestReactorMixin
from buildbot.test.util import protocols as util_protocols
from buildbot.worker.protocols import base
from buildbot.worker.protocols import msgpack


class TestListener(TestReactorMixin, unittest.TestCase):
    @defer.inlineCallbacks
    def setUp(self):
        self.setup_test_reactor()
        self.master = yield fakemaster.make_master(self)

    def test_constructor(self):
        listener = msgpack.Listener(self.master)
        self.assertEqual(listener.master, self.master)
        self.assertEqual(listener._registrations, {})

    @defer.inlineCallbacks
    def test_update_registration_simple(self):
        listener = msgpack.Listener(self.master)
        reg = yield listener.updateRegistration('example', 'pass', 'tcp:1234')
        self.assertEqual(self.master.msgmanager._registrations, [('tcp:1234', 'example', 'pass')])
        self.assertEqual(listener._registrations['example'], ('pass', 'tcp:1234', reg))

    @defer.inlineCallbacks
    def test_update_registration_pass_changed(self):
        listener = msgpack.Listener(self.master)
        listener.updateRegistration('example', 'pass', 'tcp:1234')
        reg1 = yield listener.updateRegistration('example', 'pass1', 'tcp:1234')
        self.assertEqual(listener._registrations['example'], ('pass1', 'tcp:1234', reg1))
        self.assertEqual(self.master.msgmanager._unregistrations, [('tcp:1234', 'example')])

    @defer.inlineCallbacks
    def test_update_registration_port_changed(self):
        listener = msgpack.Listener(self.master)
        listener.updateRegistration('example', 'pass', 'tcp:1234')
        reg1 = yield listener.updateRegistration('example', 'pass', 'tcp:4321')
        self.assertEqual(listener._registrations['example'], ('pass', 'tcp:4321', reg1))
        self.assertEqual(self.master.msgmanager._unregistrations, [('tcp:1234', 'example')])

    @defer.inlineCallbacks
    def test_create_connection(self):
        listener = msgpack.Listener(self.master)
        listener.before_connection_setup = mock.Mock()
        worker = mock.Mock()
        worker.workername = 'test'
        protocol = mock.Mock()

        listener.updateRegistration('example', 'pass', 'tcp:1234')
        self.master.workers.register(worker)
        conn = yield listener._create_connection(protocol, worker.workername)

        listener.before_connection_setup.assert_called_once_with(protocol, worker.workername)
        self.assertIsInstance(conn, msgpack.Connection)


class TestConnectionApi(
    util_protocols.ConnectionInterfaceTest, TestReactorMixin, unittest.TestCase
):
    @defer.inlineCallbacks
    def setUp(self):
        self.setup_test_reactor()
        self.master = yield fakemaster.make_master(self)
        self.conn = msgpack.Connection(self.master, mock.Mock(), mock.Mock())


class TestConnection(TestReactorMixin, unittest.TestCase):
    @defer.inlineCallbacks
    def setUp(self):
        self.setup_test_reactor()
        self.master = yield fakemaster.make_master(self)
        self.protocol = mock.Mock()
        self.worker = mock.Mock()
        self.worker.workername = 'test_worker'
        self.conn = msgpack.Connection(self.master, self.worker, self.protocol)

    def test_constructor(self):
        self.assertEqual(self.conn.protocol, self.protocol)
        self.assertEqual(self.conn.master, self.master)
        self.assertEqual(self.conn.worker, self.worker)

    @defer.inlineCallbacks
    def test_attached(self):
        self.conn.attached(self.protocol)
        self.worker.attached.assert_called_with(self.conn)

        self.reactor.pump([10] * 361)
        self.protocol.get_message_result.assert_called_once_with({'op': 'keepalive'})

        self.conn.detached(self.protocol)
        yield self.conn.waitShutdown()

    @defer.inlineCallbacks
    def test_detached(self):
        self.conn.attached(self.protocol)
        self.conn.detached(self.protocol)

        self.assertEqual(self.conn.keepalive_timer, None)
        self.assertEqual(self.conn.protocol, None)
        yield self.conn.waitShutdown()

    def test_lose_connection(self):
        self.conn.loseConnection()

        self.assertEqual(self.conn.keepalive_timer, None)
        self.protocol.transport.abortConnection.assert_called()

    def test_do_keepalive(self):
        self.conn._do_keepalive()
        self.protocol.get_message_result.assert_called_once_with({'op': 'keepalive'})

    @defer.inlineCallbacks
    def test_start_stop_keepalive_timer(self):
        self.conn.startKeepaliveTimer()

        self.protocol.get_message_result.assert_not_called()

        self.reactor.pump([10] * 361)
        expected_call = [
            mock.call({'op': 'keepalive'}),
        ]
        self.assertEqual(self.protocol.get_message_result.call_args_list, expected_call)

        self.reactor.pump([10] * 361)
        expected_calls = [
            mock.call({'op': 'keepalive'}),
            mock.call({'op': 'keepalive'}),
        ]
        self.assertEqual(self.protocol.get_message_result.call_args_list, expected_calls)

        self.conn.stopKeepaliveTimer()

        self.reactor.pump([10] * 361)
        expected_calls = [
            mock.call({'op': 'keepalive'}),
            mock.call({'op': 'keepalive'}),
        ]
        self.assertEqual(self.protocol.get_message_result.call_args_list, expected_calls)

        yield self.conn.waitShutdown()

    @defer.inlineCallbacks
    def test_remote_keepalive(self):
        yield self.conn.remoteKeepalive()
        self.protocol.get_message_result.assert_called_once_with({'op': 'keepalive'})

    @defer.inlineCallbacks
    def test_remote_print(self):
        yield self.conn.remotePrint(message='test')
        self.protocol.get_message_result.assert_called_once_with({'op': 'print', 'message': 'test'})

    @defer.inlineCallbacks
    def test_remote_get_worker_info(self):
        self.protocol.get_message_result.return_value = defer.succeed({'system': 'posix'})
        result = yield self.conn.remoteGetWorkerInfo()

        self.protocol.get_message_result.assert_called_once_with({'op': 'get_worker_info'})
        self.assertEqual(result, {'system': 'posix'})

    def set_up_set_builder_list(self, builders, delete_leftover_dirs=True):
        self.protocol.command_id_to_command_map = {}

        def get_message_result(*args):
            d = defer.Deferred()
            self.d_get_message_result = d
            return d

        self.protocol.get_message_result.side_effect = get_message_result
        self.conn.info = {'basedir': 'testdir'}
        self.conn.info['delete_leftover_dirs'] = delete_leftover_dirs
        self.conn.path_module = os.path
        self.conn.path_cls = PurePath
        d = self.conn.remoteSetBuilderList(builders)
        return d

    def check_message_send_response(self, command_name, args, update_msg):
        command_id = remotecommand.RemoteCommand.get_last_generated_command_id()
        self.protocol.get_message_result.assert_called_once_with({
            'op': 'start_command',
            'command_id': command_id,
            'command_name': command_name,
            'args': args,
        })
        self.protocol.get_message_result.reset_mock()
        self.d_get_message_result.callback(None)

        remote_command = self.protocol.command_id_to_command_map[command_id]
        remote_command.remote_update_msgpack(update_msg)
        remote_command.remote_complete(None)

    def check_message_set_worker_settings(self):
        newline_re = r'(\r\n|\r(?=.)|\033\[u|\033\[[0-9]+;[0-9]+[Hf]|\033\[2J|\x08+)'
        self.protocol.get_message_result.assert_called_once_with({
            'op': 'set_worker_settings',
            'args': {
                'newline_re': newline_re,
                'max_line_length': 4096,
                'buffer_timeout': 5,
                'buffer_size': 64 * 1024,
            },
        })
        self.protocol.get_message_result.reset_mock()
        self.d_get_message_result.callback(None)

    @defer.inlineCallbacks
    def test_remote_set_builder_list_no_rmdir(self):
        d = self.set_up_set_builder_list([('builder1', 'test_dir1'), ('builder2', 'test_dir2')])
        self.check_message_set_worker_settings()

        self.check_message_send_response(
            'listdir', {'path': 'testdir'}, [('files', ['dir1', 'dir2', 'dir3']), ('rc', 0)]
        )

        path = os.path.join('testdir', 'dir1')
        self.check_message_send_response('stat', {'path': path}, [('stat', (1,)), ('rc', 0)])

        path = os.path.join('testdir', 'dir2')
        self.check_message_send_response('stat', {'path': path}, [('stat', (1,)), ('rc', 0)])

        path = os.path.join('testdir', 'dir3')
        self.check_message_send_response('stat', {'path': path}, [('stat', (1,)), ('rc', 0)])

        paths = [
            os.path.join('testdir', 'info'),
            os.path.join('testdir', 'test_dir1'),
            os.path.join('testdir', 'test_dir2'),
        ]
        self.check_message_send_response('mkdir', {'paths': paths}, [('rc', 0)])

        r = yield d
        self.assertEqual(r, ['builder1', 'builder2'])
        self.protocol.get_message_result.assert_not_called()

    @defer.inlineCallbacks
    def test_remote_set_builder_list_do_rmdir(self):
        d = self.set_up_set_builder_list([('builder1', 'test_dir1'), ('builder2', 'test_dir2')])
        self.check_message_set_worker_settings()

        self.check_message_send_response(
            'listdir', {'path': 'testdir'}, [('files', ['dir1', 'dir2', 'dir3']), ('rc', 0)]
        )

        path = os.path.join('testdir', 'dir1')
        self.check_message_send_response(
            'stat', {'path': path}, [('stat', (stat.S_IFDIR,)), ('rc', 0)]
        )

        path = os.path.join('testdir', 'dir2')
        self.check_message_send_response(
            'stat', {'path': path}, [('stat', (stat.S_IFDIR,)), ('rc', 0)]
        )

        path = os.path.join('testdir', 'dir3')
        self.check_message_send_response(
            'stat', {'path': path}, [('stat', (stat.S_IFDIR,)), ('rc', 0)]
        )

        paths = [
            os.path.join('testdir', 'dir1'),
            os.path.join('testdir', 'dir2'),
            os.path.join('testdir', 'dir3'),
        ]
        self.check_message_send_response('rmdir', {'paths': paths}, [('rc', 0)])

        paths = [
            os.path.join('testdir', 'info'),
            os.path.join('testdir', 'test_dir1'),
            os.path.join('testdir', 'test_dir2'),
        ]
        self.check_message_send_response('mkdir', {'paths': paths}, [('rc', 0)])

        r = yield d
        self.assertEqual(r, ['builder1', 'builder2'])
        self.protocol.get_message_result.assert_not_called()

    @defer.inlineCallbacks
    def test_remote_set_builder_list_no_rmdir_leave_leftover_dirs(self):
        d = self.set_up_set_builder_list(
            [('builder1', 'test_dir1'), ('builder2', 'test_dir2')], delete_leftover_dirs=False
        )
        self.check_message_set_worker_settings()

        self.check_message_send_response(
            'listdir', {'path': 'testdir'}, [('files', ['dir1', 'dir2', 'dir3']), ('rc', 0)]
        )

        paths = [
            os.path.join('testdir', 'info'),
            os.path.join('testdir', 'test_dir1'),
            os.path.join('testdir', 'test_dir2'),
        ]
        self.check_message_send_response('mkdir', {'paths': paths}, [('rc', 0)])

        r = yield d
        self.assertEqual(r, ['builder1', 'builder2'])
        self.protocol.get_message_result.assert_not_called()

    @defer.inlineCallbacks
    def test_remote_set_builder_list_no_mkdir_from_files(self):
        d = self.set_up_set_builder_list([('builder1', 'test_dir1'), ('builder2', 'test_dir2')])
        self.check_message_set_worker_settings()

        self.check_message_send_response(
            'listdir', {'path': 'testdir'}, [('files', ['dir1', 'test_dir2']), ('rc', 0)]
        )

        path = os.path.join('testdir', 'dir1')
        self.check_message_send_response('stat', {'path': path}, [('stat', (1,)), ('rc', 0)])

        paths = [os.path.join('testdir', 'info'), os.path.join('testdir', 'test_dir1')]
        self.check_message_send_response('mkdir', {'paths': paths}, [('rc', 0)])

        r = yield d
        self.assertEqual(r, ['builder1', 'builder2'])
        self.protocol.get_message_result.assert_not_called()

    @defer.inlineCallbacks
    def test_remote_set_builder_list_no_mkdir(self):
        d = self.set_up_set_builder_list([('builder1', 'test_dir1'), ('builder2', 'test_dir2')])
        self.check_message_set_worker_settings()

        self.check_message_send_response(
            'listdir',
            {'path': 'testdir'},
            [('files', ['test_dir1', 'test_dir2', 'info']), ('rc', 0)],
        )

        r = yield d
        self.assertEqual(r, ['builder1', 'builder2'])
        self.protocol.get_message_result.assert_not_called()

    @defer.inlineCallbacks
    def test_remote_set_builder_list_key_is_missing(self):
        d = self.set_up_set_builder_list([('builder1', 'test_dir1'), ('builder2', 'test_dir2')])
        self.check_message_set_worker_settings()

        self.check_message_send_response(
            'listdir', {'path': 'testdir'}, [('no_key', []), ('rc', 0)]
        )

        with self.assertRaisesRegex(Exception, "Key 'files' is missing."):
            yield d

        self.protocol.get_message_result.assert_not_called()

    @defer.inlineCallbacks
    def test_remote_set_builder_list_key_rc_not_zero(self):
        d = self.set_up_set_builder_list([('builder1', 'test_dir1'), ('builder2', 'test_dir2')])
        self.check_message_set_worker_settings()

        self.check_message_send_response('listdir', {'path': 'testdir'}, [('rc', 123)])

        with self.assertRaisesRegex(Exception, "Error number: 123"):
            yield d

        self.protocol.get_message_result.assert_not_called()

    @parameterized.expand([
        ('want_stdout', 0, False),
        ('want_stdout', 1, True),
        ('want_stderr', 0, False),
        ('want_stderr', 1, True),
        (None, None, None),
    ])
    @defer.inlineCallbacks
    def test_remote_start_command_args_update(self, arg_name, arg_value, expected_value):
        self.protocol.get_message_result.return_value = defer.succeed(None)

        rc_instance = base.RemoteCommandImpl()
        result_command_id_to_command_map = {1: rc_instance}
        self.protocol.command_id_to_command_map = {}
        args = {'args': 'args'}

        if arg_name is not None:
            args[arg_name] = arg_value

        yield self.conn.remoteStartCommand(rc_instance, 'builder', 1, 'command', args)
        expected_args = args.copy()
        if arg_name is not None:
            expected_args[arg_name] = expected_value

        self.assertEqual(result_command_id_to_command_map, self.protocol.command_id_to_command_map)
        self.protocol.get_message_result.assert_called_with({
            'op': 'start_command',
            'builder_name': 'builder',
            'command_id': 1,
            'command_name': 'command',
            'args': expected_args,
        })

    @defer.inlineCallbacks
    def test_remote_shutdown(self):
        self.protocol.get_message_result.return_value = defer.succeed(None)
        yield self.conn.remoteShutdown()

        self.protocol.get_message_result.assert_called_once_with({'op': 'shutdown'})

    @defer.inlineCallbacks
    def test_remote_interrupt_command(self):
        self.protocol.get_message_result.return_value = defer.succeed(None)
        yield self.conn.remoteInterruptCommand('builder', 1, 'test')

        self.protocol.get_message_result.assert_called_once_with({
            'op': 'interrupt_command',
            'builder_name': 'builder',
            'command_id': 1,
            'why': 'test',
        })

    def test_perspective_keepalive(self):
        self.conn.perspective_keepalive()
        self.conn.worker.messageReceivedFromWorker.assert_called_once_with()

    def test_perspective_shutdown(self):
        self.conn.perspective_shutdown()

        self.conn.worker.shutdownRequested.assert_called_once_with()
        self.conn.worker.messageReceivedFromWorker.assert_called_once_with()



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from twisted.internet import defer
from twisted.trial import unittest

from buildbot.db import projects
from buildbot.test import fakedb
from buildbot.test.fake import fakemaster
from buildbot.test.reactor import TestReactorMixin


def project_key(builder):
    return builder.id


class Tests(TestReactorMixin, unittest.TestCase):
    @defer.inlineCallbacks
    def setUp(self):
        self.setup_test_reactor()
        self.master = yield fakemaster.make_master(self, wantDb=True)
        self.db = self.master.db

    @defer.inlineCallbacks
    def test_update_project_info(self):
        yield self.db.insert_test_data([
            fakedb.Project(id=7, name='fake_project7'),
        ])

        yield self.db.projects.update_project_info(
            7, "slug7", "project7 desc", "format", "html desc"
        )
        dbdict = yield self.db.projects.get_project(7)
        self.assertIsInstance(dbdict, projects.ProjectModel)
        self.assertEqual(
            dbdict,
            projects.ProjectModel(
                id=7,
                name="fake_project7",
                slug="slug7",
                description="project7 desc",
                description_format="format",
                description_html="html desc",
            ),
        )

    @defer.inlineCallbacks
    def test_find_project_id_new(self):
        id = yield self.db.projects.find_project_id('fake_project')
        dbdict = yield self.db.projects.get_project(id)
        self.assertEqual(
            dbdict,
            projects.ProjectModel(
                id=id,
                name="fake_project",
                slug="fake_project",
                description=None,
                description_format=None,
                description_html=None,
            ),
        )

    @defer.inlineCallbacks
    def test_find_project_id_new_no_auto_create(self):
        id = yield self.db.projects.find_project_id('fake_project', auto_create=False)
        self.assertIsNone(id)

    @defer.inlineCallbacks
    def test_find_project_id_exists(self):
        yield self.db.insert_test_data([
            fakedb.Project(id=7, name='fake_project'),
        ])
        id = yield self.db.projects.find_project_id('fake_project')
        self.assertEqual(id, 7)

    @defer.inlineCallbacks
    def test_get_project(self):
        yield self.db.insert_test_data([
            fakedb.Project(id=7, name='fake_project'),
        ])
        dbdict = yield self.db.projects.get_project(7)
        self.assertIsInstance(dbdict, projects.ProjectModel)
        self.assertEqual(
            dbdict,
            projects.ProjectModel(
                id=7,
                name="fake_project",
                slug="fake_project",
                description=None,
                description_format=None,
                description_html=None,
            ),
        )

    @defer.inlineCallbacks
    def test_get_project_missing(self):
        dbdict = yield self.db.projects.get_project(7)
        self.assertIsNone(dbdict)

    @defer.inlineCallbacks
    def test_get_projects(self):
        yield self.db.insert_test_data([
            fakedb.Project(id=7, name="fake_project7"),
            fakedb.Project(id=8, name="fake_project8"),
            fakedb.Project(id=9, name="fake_project9"),
        ])
        dblist = yield self.db.projects.get_projects()
        for dbdict in dblist:
            self.assertIsInstance(dbdict, projects.ProjectModel)
        self.assertEqual(
            sorted(dblist, key=project_key),
            sorted(
                [
                    projects.ProjectModel(
                        id=7,
                        name="fake_project7",
                        slug="fake_project7",
                        description=None,
                        description_format=None,
                        description_html=None,
                    ),
                    projects.ProjectModel(
                        id=8,
                        name="fake_project8",
                        slug="fake_project8",
                        description=None,
                        description_format=None,
                        description_html=None,
                    ),
                    projects.ProjectModel(
                        id=9,
                        name="fake_project9",
                        slug="fake_project9",
                        description=None,
                        description_format=None,
                        description_html=None,
                    ),
                ],
                key=project_key,
            ),
        )

    @defer.inlineCallbacks
    def test_get_projects_empty(self):
        dblist = yield self.db.projects.get_projects()
        self.assertEqual(dblist, [])

    @defer.inlineCallbacks
    def test_get_active_projects(self):
        yield self.db.insert_test_data([
            fakedb.Project(id=1, name='fake_project1'),
            fakedb.Project(id=2, name='fake_project2'),
            fakedb.Project(id=3, name='fake_project3'),
            fakedb.Master(id=100),
            fakedb.Builder(id=200, name="builder_200", projectid=2),
            fakedb.Builder(id=201, name="builder_201", projectid=3),
            fakedb.BuilderMaster(id=300, builderid=200, masterid=100),
        ])
        dblist = yield self.db.projects.get_active_projects()
        for dbdict in dblist:
            self.assertIsInstance(dbdict, projects.ProjectModel)
        self.assertEqual(
            dblist,
            [
                projects.ProjectModel(
                    id=2,
                    name="fake_project2",
                    slug="fake_project2",
                    description=None,
                    description_format=None,
                    description_html=None,
                )
            ],
        )



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from __future__ import annotations

from typing import TYPE_CHECKING
from unittest import mock

from twisted.internet import defer
from twisted.trial import unittest

from buildbot import config
from buildbot.errors import CaptureCallbackError
from buildbot.statistics import capture
from buildbot.statistics import stats_service
from buildbot.statistics import storage_backends
from buildbot.statistics.storage_backends.base import StatsStorageBase
from buildbot.statistics.storage_backends.influxdb_client import InfluxStorageService
from buildbot.test import fakedb
from buildbot.test.fake import fakemaster
from buildbot.test.fake import fakestats
from buildbot.test.reactor import TestReactorMixin
from buildbot.test.steps import TestBuildStepMixin
from buildbot.test.util import logging

if TYPE_CHECKING:
    from buildbot.db.builds import BuildModel


class TestStatsServicesBase(TestReactorMixin, unittest.TestCase):
    BUILDER_NAMES = ['builder1', 'builder2']
    BUILDER_IDS = [1, 2]

    @defer.inlineCallbacks
    def setUp(self):
        self.setup_test_reactor()
        self.master = yield fakemaster.make_master(self, wantMq=True, wantData=True, wantDb=True)

        yield self.master.db.insert_test_data(
            [
                fakedb.Master(id=fakedb.FakeDBConnector.MASTER_ID),
            ]
            + [
                fakedb.Builder(id=builderid, name=name)
                for builderid, name in zip(self.BUILDER_IDS, self.BUILDER_NAMES)
            ]
        )

        self.stats_service = stats_service.StatsService(
            storage_backends=[fakestats.FakeStatsStorageService()], name="FakeStatsService"
        )
        yield self.stats_service.setServiceParent(self.master)
        yield self.master.startService()
        self.addCleanup(self.master.stopService)


class TestStatsServicesConfiguration(TestStatsServicesBase):
    @defer.inlineCallbacks
    def test_reconfig_with_no_storage_backends(self):
        new_storage_backends = []
        yield self.stats_service.reconfigService(new_storage_backends)
        self.checkEqual(new_storage_backends)

    @defer.inlineCallbacks
    def test_reconfig_with_fake_storage_backend(self):
        new_storage_backends = [
            fakestats.FakeStatsStorageService(name='One'),
            fakestats.FakeStatsStorageService(name='Two'),
        ]
        yield self.stats_service.reconfigService(new_storage_backends)
        self.checkEqual(new_storage_backends)

    @defer.inlineCallbacks
    def test_reconfig_with_consumers(self):
        backend = fakestats.FakeStatsStorageService(name='One')
        backend.captures = [capture.CaptureProperty('test_builder', 'test')]
        new_storage_backends = [backend]

        yield self.stats_service.reconfigService(new_storage_backends)
        yield self.stats_service.reconfigService(new_storage_backends)
        self.assertEqual(len(self.master.mq.qrefs), 1)

    @defer.inlineCallbacks
    def test_bad_configuration(self):
        # Reconfigure with a bad configuration.
        new_storage_backends = [mock.Mock()]
        with self.assertRaises(TypeError):
            yield self.stats_service.reconfigService(new_storage_backends)

    def checkEqual(self, new_storage_backends):
        # Check whether the new_storage_backends was set in reconfigService
        registeredStorageServices = [
            s
            for s in self.stats_service.registeredStorageServices
            if isinstance(s, StatsStorageBase)
        ]
        for s in new_storage_backends:
            if s not in registeredStorageServices:
                raise AssertionError("reconfigService failed. Not all storage services registered.")


class TestInfluxDB(TestStatsServicesBase, logging.LoggingMixin):
    # Smooth test of influx db service. We don't want to force people to install influxdb, so we
    # just disable this unit test if the influxdb module is not installed,
    # using SkipTest

    @defer.inlineCallbacks
    def test_influxdb_not_installed(self):
        captures = [capture.CaptureProperty('test_builder', 'test')]
        try:
            # Try to import
            import influxdb  # pylint: disable=import-outside-toplevel

            # consume it somehow to please pylint
            _ = influxdb
        except ImportError:
            with self.assertRaises(config.ConfigErrors):
                InfluxStorageService(
                    "fake_url", 12345, "fake_user", "fake_password", "fake_db", captures
                )

        # if instead influxdb is installed, then initialize it - no errors
        # should be realized
        else:
            new_storage_backends = [
                InfluxStorageService(
                    "fake_url", 12345, "fake_user", "fake_password", "fake_db", captures
                )
            ]
            yield self.stats_service.reconfigService(new_storage_backends)

    @defer.inlineCallbacks
    def test_influx_storage_service_fake_install(self):
        # use a fake InfluxDBClient to test InfluxStorageService in systems which
        # don't have influxdb installed. Primarily useful for test coverage.
        self.patch(storage_backends.influxdb_client, 'InfluxDBClient', fakestats.FakeInfluxDBClient)
        captures = [capture.CaptureProperty('test_builder', 'test')]
        new_storage_backends = [
            InfluxStorageService(
                "fake_url", "fake_port", "fake_user", "fake_password", "fake_db", captures
            )
        ]
        yield self.stats_service.reconfigService(new_storage_backends)

    def test_influx_storage_service_post_value(self):
        # test the thd_postStatsValue method of InfluxStorageService
        self.patch(storage_backends.influxdb_client, 'InfluxDBClient', fakestats.FakeInfluxDBClient)
        svc = InfluxStorageService(
            "fake_url", "fake_port", "fake_user", "fake_password", "fake_db", "fake_stats"
        )
        post_data = {'name': 'test', 'value': 'test'}
        context = {'x': 'y'}
        svc.thd_postStatsValue(post_data, "test_series_name", context)
        data = {
            'measurement': "test_series_name",
            'fields': {"name": "test", "value": "test"},
            'tags': {'x': 'y'},
        }
        points = [data]
        self.assertEqual(svc.client.points, points)

    def test_influx_service_not_inited(self):
        self.setUpLogging()
        self.patch(storage_backends.influxdb_client, 'InfluxDBClient', fakestats.FakeInfluxDBClient)
        svc = InfluxStorageService(
            "fake_url", "fake_port", "fake_user", "fake_password", "fake_db", "fake_stats"
        )
        svc._inited = False
        svc.thd_postStatsValue("test", "test", "test")
        self.assertLogged("Service.*not initialized")


class TestStatsServicesConsumers(TestBuildStepMixin, TestStatsServicesBase):
    """
    Test the stats service from a fake step
    """

    @defer.inlineCallbacks
    def setUp(self):
        yield super().setUp()
        self.routingKey = ("builders", self.BUILDER_IDS[0], "builds", 1, "finished")
        self.master.mq.verifyMessages = False

    @defer.inlineCallbacks
    def setupBuild(self):
        yield self.master.db.insert_test_data([
            fakedb.Worker(id=1, name='wrk'),
            fakedb.Buildset(id=8822),
            fakedb.BuildRequest(
                id=1,
                buildsetid=8822,
                builderid=self.BUILDER_IDS[0],
            ),
            fakedb.Build(
                id=1,
                masterid=fakedb.FakeDBConnector.MASTER_ID,
                workerid=1,
                builderid=self.BUILDER_IDS[0],
                buildrequestid=1,
                number=1,
            ),
        ])

    @defer.inlineCallbacks
    def setupFakeStorage(self, captures):
        self.fake_storage_service = fakestats.FakeStatsStorageService()
        self.fake_storage_service.captures = captures
        yield self.stats_service.reconfigService([self.fake_storage_service])

    def get_dict(self, build: BuildModel):
        return {
            "buildid": 1,
            "number": build.number,
            "builderid": build.builderid,
            "buildrequestid": build.buildrequestid,
            "workerid": build.workerid,
            "masterid": build.masterid,
            "started_at": build.started_at,
            "complete": True,
            "complete_at": build.complete_at,
            "state_string": '',
            "results": 0,
        }

    @defer.inlineCallbacks
    def end_build_call_consumers(self):
        self.master.db.builds.finishBuild(buildid=1, results=0)
        build = yield self.master.db.builds.getBuild(buildid=1)
        self.master.mq.callConsumer(self.routingKey, self.get_dict(build))

    @defer.inlineCallbacks
    def test_property_capturing(self):
        self.setupFakeStorage([capture.CaptureProperty('builder1', 'test_name')])
        yield self.setupBuild()
        self.master.db.builds.setBuildProperty(1, 'test_name', 'test_value', 'test_source')
        yield self.end_build_call_consumers()

        self.assertEqual(
            [
                (
                    {'name': 'test_name', 'value': 'test_value'},
                    'builder1-test_name',
                    {'build_number': '1', 'builder_name': 'builder1'},
                )
            ],
            self.fake_storage_service.stored_data,
        )

    @defer.inlineCallbacks
    def test_property_capturing_all_builders(self):
        self.setupFakeStorage([capture.CapturePropertyAllBuilders('test_name')])
        yield self.setupBuild()
        self.master.db.builds.setBuildProperty(1, 'test_name', 'test_value', 'test_source')
        yield self.end_build_call_consumers()

        self.assertEqual(
            [
                (
                    {'name': 'test_name', 'value': 'test_value'},
                    'builder1-test_name',
                    {'build_number': '1', 'builder_name': 'builder1'},
                )
            ],
            self.fake_storage_service.stored_data,
        )

    @defer.inlineCallbacks
    def test_property_capturing_regex(self):
        self.setupFakeStorage([capture.CaptureProperty('builder1', 'test_n.*', regex=True)])
        yield self.setupBuild()
        self.master.db.builds.setBuildProperty(1, 'test_name', 'test_value', 'test_source')
        yield self.end_build_call_consumers()

        self.assertEqual(
            [
                (
                    {'name': 'test_name', 'value': 'test_value'},
                    'builder1-test_name',
                    {'build_number': '1', 'builder_name': 'builder1'},
                )
            ],
            self.fake_storage_service.stored_data,
        )

    @defer.inlineCallbacks
    def test_property_capturing_error(self):
        self.setupFakeStorage([capture.CaptureProperty('builder1', 'test')])
        yield self.setupBuild()
        self.master.db.builds.setBuildProperty(1, 'test_name', 'test_value', 'test_source')
        self.master.db.builds.finishBuild(buildid=1, results=0)
        build = yield self.master.db.builds.getBuild(buildid=1)
        cap = self.fake_storage_service.captures[0]
        with self.assertRaises(CaptureCallbackError):
            yield cap.consume(self.routingKey, self.get_dict(build))

    @defer.inlineCallbacks
    def test_property_capturing_alt_callback(self):
        def cb(*args, **kwargs):
            return 'test_value'

        self.setupFakeStorage([capture.CaptureProperty('builder1', 'test_name', cb)])
        yield self.setupBuild()
        self.master.db.builds.setBuildProperty(1, 'test_name', 'test_value', 'test_source')
        yield self.end_build_call_consumers()

        self.assertEqual(
            [
                (
                    {'name': 'test_name', 'value': 'test_value'},
                    'builder1-test_name',
                    {'build_number': '1', 'builder_name': 'builder1'},
                )
            ],
            self.fake_storage_service.stored_data,
        )

    @defer.inlineCallbacks
    def test_build_start_time_capturing(self):
        self.setupFakeStorage([capture.CaptureBuildStartTime('builder1')])
        yield self.setupBuild()
        yield self.end_build_call_consumers()

        self.assertEqual(
            'start-time', next(iter(self.fake_storage_service.stored_data[0][0].keys()))
        )

    @defer.inlineCallbacks
    def test_build_start_time_capturing_all_builders(self):
        self.setupFakeStorage([capture.CaptureBuildStartTimeAllBuilders()])
        yield self.setupBuild()
        yield self.end_build_call_consumers()

        self.assertEqual(
            'start-time', next(iter(self.fake_storage_service.stored_data[0][0].keys()))
        )

    @defer.inlineCallbacks
    def test_build_start_time_capturing_alt_callback(self):
        def cb(*args, **kwargs):
            return '2015-07-08T01:45:17.391018'

        self.setupFakeStorage([capture.CaptureBuildStartTime('builder1', cb)])
        yield self.setupBuild()
        yield self.end_build_call_consumers()

        self.assertEqual(
            'start-time', next(iter(self.fake_storage_service.stored_data[0][0].keys()))
        )

    @defer.inlineCallbacks
    def test_build_end_time_capturing(self):
        self.setupFakeStorage([capture.CaptureBuildEndTime('builder1')])
        yield self.setupBuild()
        yield self.end_build_call_consumers()

        self.assertEqual('end-time', next(iter(self.fake_storage_service.stored_data[0][0].keys())))

    @defer.inlineCallbacks
    def test_build_end_time_capturing_all_builders(self):
        self.setupFakeStorage([capture.CaptureBuildEndTimeAllBuilders()])
        yield self.setupBuild()
        yield self.end_build_call_consumers()

        self.assertEqual('end-time', next(iter(self.fake_storage_service.stored_data[0][0].keys())))

    @defer.inlineCallbacks
    def test_build_end_time_capturing_alt_callback(self):
        def cb(*args, **kwargs):
            return '2015-07-08T01:45:17.391018'

        self.setupFakeStorage([capture.CaptureBuildEndTime('builder1', cb)])
        yield self.setupBuild()
        yield self.end_build_call_consumers()

        self.assertEqual('end-time', next(iter(self.fake_storage_service.stored_data[0][0].keys())))

    @defer.inlineCallbacks
    def build_time_capture_helper(self, time_type, cb=None):
        self.setupFakeStorage([
            capture.CaptureBuildDuration('builder1', report_in=time_type, callback=cb)
        ])
        yield self.setupBuild()
        yield self.end_build_call_consumers()

    @defer.inlineCallbacks
    def test_build_duration_capturing_seconds(self):
        yield self.build_time_capture_helper('seconds')
        self.assertEqual('duration', next(iter(self.fake_storage_service.stored_data[0][0].keys())))

    @defer.inlineCallbacks
    def test_build_duration_capturing_minutes(self):
        yield self.build_time_capture_helper('minutes')
        self.assertEqual('duration', next(iter(self.fake_storage_service.stored_data[0][0].keys())))

    @defer.inlineCallbacks
    def test_build_duration_capturing_hours(self):
        yield self.build_time_capture_helper('hours')
        self.assertEqual('duration', next(iter(self.fake_storage_service.stored_data[0][0].keys())))

    def test_build_duration_report_in_error(self):
        with self.assertRaises(config.ConfigErrors):
            capture.CaptureBuildDuration('builder1', report_in='foobar')

    @defer.inlineCallbacks
    def test_build_duration_capturing_alt_callback(self):
        def cb(*args, **kwargs):
            return 10

        yield self.build_time_capture_helper('seconds', cb)
        self.assertEqual('duration', next(iter(self.fake_storage_service.stored_data[0][0].keys())))

    @defer.inlineCallbacks
    def test_build_duration_capturing_all_builders(self):
        self.setupFakeStorage([capture.CaptureBuildDurationAllBuilders()])
        yield self.setupBuild()
        yield self.end_build_call_consumers()

        self.assertEqual('duration', next(iter(self.fake_storage_service.stored_data[0][0].keys())))

    @defer.inlineCallbacks
    def test_build_times_capturing_error(self):
        def cb(*args, **kwargs):
            raise TypeError

        self.setupFakeStorage([capture.CaptureBuildStartTime('builder1', cb)])
        yield self.setupBuild()
        self.master.db.builds.setBuildProperty(1, 'test_name', 'test_value', 'test_source')
        self.master.db.builds.finishBuild(buildid=1, results=0)
        build = yield self.master.db.builds.getBuild(buildid=1)
        cap = self.fake_storage_service.captures[0]
        with self.assertRaises(CaptureCallbackError):
            yield cap.consume(self.routingKey, self.get_dict(build))

        self.setupFakeStorage([capture.CaptureBuildEndTime('builder1', cb)])
        cap = self.fake_storage_service.captures[0]

        with self.assertRaises(CaptureCallbackError):
            yield cap.consume(self.routingKey, self.get_dict(build))

        self.setupFakeStorage([capture.CaptureBuildDuration('builder1', callback=cb)])
        cap = self.fake_storage_service.captures[0]

        with self.assertRaises(CaptureCallbackError):
            yield cap.consume(self.routingKey, self.get_dict(build))

    @defer.inlineCallbacks
    def test_yield_metrics_value(self):
        self.setupFakeStorage([capture.CaptureBuildStartTime('builder1')])
        yield self.setupBuild()
        yield self.end_build_call_consumers()
        yield self.stats_service.yieldMetricsValue('test', {'test': 'test'}, 1)

        build_data = yield self.stats_service.master.data.get(('builds', 1))
        routingKey = ("stats-yieldMetricsValue", "stats-yield-data")

        msg = {'data_name': 'test', 'post_data': {'test': 'test'}, 'build_data': build_data}

        exp = [(routingKey, msg)]
        self.stats_service.master.mq.assertProductions(exp)

    @defer.inlineCallbacks
    def test_capture_data(self):
        self.setupFakeStorage([capture.CaptureData('test', 'builder1')])
        yield self.setupBuild()
        self.master.db.builds.finishBuild(buildid=1, results=0)
        build_data = yield self.stats_service.master.data.get(('builds', 1))

        msg = {'data_name': 'test', 'post_data': {'test': 'test'}, 'build_data': build_data}

        routingKey = ("stats-yieldMetricsValue", "stats-yield-data")
        self.master.mq.callConsumer(routingKey, msg)
        self.assertEqual(
            [
                (
                    {'test': 'test'},
                    'builder1-test',
                    {'build_number': '1', 'builder_name': 'builder1'},
                )
            ],
            self.fake_storage_service.stored_data,
        )

    @defer.inlineCallbacks
    def test_capture_data_all_builders(self):
        self.setupFakeStorage([capture.CaptureDataAllBuilders('test')])
        yield self.setupBuild()
        self.master.db.builds.finishBuild(buildid=1, results=0)
        build_data = yield self.stats_service.master.data.get(('builds', 1))

        msg = {'data_name': 'test', 'post_data': {'test': 'test'}, 'build_data': build_data}

        routingKey = ("stats-yieldMetricsValue", "stats-yield-data")
        self.master.mq.callConsumer(routingKey, msg)
        self.assertEqual(
            [
                (
                    {'test': 'test'},
                    'builder1-test',
                    {'build_number': '1', 'builder_name': 'builder1'},
                )
            ],
            self.fake_storage_service.stored_data,
        )

    @defer.inlineCallbacks
    def test_capture_data_alt_callback(self):
        def cb(*args, **kwargs):
            return {'test': 'test'}

        self.setupFakeStorage([capture.CaptureData('test', 'builder1', cb)])
        yield self.setupBuild()
        self.master.db.builds.finishBuild(buildid=1, results=0)
        build_data = yield self.stats_service.master.data.get(('builds', 1))

        msg = {'data_name': 'test', 'post_data': {'test': 'test'}, 'build_data': build_data}

        routingKey = ("stats-yieldMetricsValue", "stats-yield-data")
        self.master.mq.callConsumer(routingKey, msg)
        self.assertEqual(
            [
                (
                    {'test': 'test'},
                    'builder1-test',
                    {'build_number': '1', 'builder_name': 'builder1'},
                )
            ],
            self.fake_storage_service.stored_data,
        )

    @defer.inlineCallbacks
    def test_capture_data_error(self):
        def cb(*args, **kwargs):
            raise TypeError

        self.setupFakeStorage([capture.CaptureData('test', 'builder1', cb)])
        yield self.setupBuild()
        self.master.db.builds.finishBuild(buildid=1, results=0)
        build_data = yield self.stats_service.master.data.get(('builds', 1))

        msg = {'data_name': 'test', 'post_data': {'test': 'test'}, 'build_data': build_data}

        routingKey = ("stats-yieldMetricsValue", "stats-yield-data")
        cap = self.fake_storage_service.captures[0]
        with self.assertRaises(CaptureCallbackError):
            yield cap.consume(routingKey, msg)



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from unittest import mock

from parameterized import parameterized
from twisted.internet import defer
from twisted.trial import unittest

from buildbot.data import projects
from buildbot.data import resultspec
from buildbot.db.projects import ProjectModel
from buildbot.test import fakedb
from buildbot.test.fake import fakemaster
from buildbot.test.reactor import TestReactorMixin
from buildbot.test.util import endpoint
from buildbot.test.util import interfaces


class ProjectEndpoint(endpoint.EndpointMixin, unittest.TestCase):
    endpointClass = projects.ProjectEndpoint
    resourceTypeClass = projects.Project

    @defer.inlineCallbacks
    def setUp(self):
        yield self.setUpEndpoint()
        yield self.master.db.insert_test_data([
            fakedb.Project(id=1, name='project1'),
            fakedb.Project(id=2, name='project2'),
        ])

    @defer.inlineCallbacks
    def test_get_existing_id(self):
        project = yield self.callGet(('projects', 2))

        self.validateData(project)
        self.assertEqual(project['name'], 'project2')

    @defer.inlineCallbacks
    def test_get_existing_name(self):
        project = yield self.callGet(('projects', 'project2'))

        self.validateData(project)
        self.assertEqual(project['name'], 'project2')

    @defer.inlineCallbacks
    def test_get_missing(self):
        project = yield self.callGet(('projects', 99))

        self.assertIsNone(project)

    @defer.inlineCallbacks
    def test_get_missing_with_name(self):
        project = yield self.callGet(('projects', 'project99'))

        self.assertIsNone(project)


class ProjectsEndpoint(endpoint.EndpointMixin, unittest.TestCase):
    endpointClass = projects.ProjectsEndpoint
    resourceTypeClass = projects.Project

    @defer.inlineCallbacks
    def setUp(self):
        yield self.setUpEndpoint()
        yield self.master.db.insert_test_data([
            fakedb.Project(id=1, name='project1'),
            fakedb.Project(id=2, name='project2'),
            fakedb.Project(id=3, name='project3'),
            fakedb.Master(id=100),
            fakedb.Builder(id=200, projectid=2),
            fakedb.Builder(id=201, projectid=3),
            fakedb.BuilderMaster(id=300, builderid=200, masterid=100),
        ])

    @parameterized.expand([
        ('no_filter', None, [1, 2, 3]),
        ('active', True, [2]),
        ('inactive', False, [1, 3]),
    ])
    @defer.inlineCallbacks
    def test_get(self, name, active_filter, expected_projectids):
        result_spec = None
        if active_filter is not None:
            result_spec = resultspec.OptimisedResultSpec(
                filters=[resultspec.Filter('active', 'eq', [active_filter])]
            )

        projects = yield self.callGet(('projects',), resultSpec=result_spec)

        for b in projects:
            self.validateData(b)

        self.assertEqual(sorted([b['projectid'] for b in projects]), expected_projectids)


class Project(interfaces.InterfaceTests, TestReactorMixin, unittest.TestCase):
    @defer.inlineCallbacks
    def setUp(self):
        self.setup_test_reactor()
        self.master = yield fakemaster.make_master(self, wantMq=True, wantDb=True, wantData=True)
        self.rtype = projects.Project(self.master)
        yield self.master.db.insert_test_data([
            fakedb.Project(id=13, name="fake_project"),
        ])

    def test_signature_find_project_id(self):
        @self.assertArgSpecMatches(
            self.master.data.updates.find_project_id,  # fake
            self.rtype.find_project_id,
        )  # real
        def find_project_id(self, name: str, auto_create: bool = True):
            pass

    def test_find_project_id(self):
        # this just passes through to the db method, so test that
        rv = defer.succeed(None)
        self.master.db.projects.find_project_id = mock.Mock(return_value=rv)
        self.assertIdentical(self.rtype.find_project_id('foo'), rv)

    def test_signature_update_project_info(self):
        @self.assertArgSpecMatches(self.master.data.updates.update_project_info)
        def update_project_info(
            self, projectid, slug, description, description_format, description_html
        ):
            pass

    @defer.inlineCallbacks
    def test_update_project_info(self):
        yield self.master.data.updates.update_project_info(
            13,
            "slug13",
            "project13 desc",
            "format",
            "html desc",
        )
        projects = yield self.master.db.projects.get_projects()
        self.assertEqual(
            projects,
            [
                ProjectModel(
                    id=13,
                    name="fake_project",
                    slug="slug13",
                    description="project13 desc",
                    description_format="format",
                    description_html="html desc",
                )
            ],
        )



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from twisted.trial.unittest import TestCase

from buildbot.process.builder import Builder
from buildbot.process.workerforbuilder import AbstractWorkerForBuilder
from buildbot.worker.base import AbstractWorker


class TestAbstractWorkerForBuilder(TestCase):
    """
    Tests for ``AbstractWorkerForBuilder``.
    """

    def test_buildStarted_called(self):
        """
        If the worker associated to worker builder has a ``buildStarted`` method,
        calling ``buildStarted`` on the worker builder calls the method on the
        worker with the workerforbuilder as an argument.
        """

        class ConcreteWorker(AbstractWorker):
            _buildStartedCalls = []

            def buildStarted(self, workerforbuilder):
                self._buildStartedCalls.append(workerforbuilder)

        fake_builder = Builder("fake_builder")
        worker = ConcreteWorker("worker", "pass")
        workerforbuilder = AbstractWorkerForBuilder(fake_builder)
        # FIXME: This should call attached, instead of setting the attribute
        # directly
        workerforbuilder.worker = worker
        workerforbuilder.buildStarted()

        self.assertEqual(ConcreteWorker._buildStartedCalls, [workerforbuilder])

    def test_buildStarted_missing(self):
        """
        If the worker associated to worker builder doesn't not have a
        ``buildStarted`` method, calling ``buildStarted`` on the worker builder
        doesn't raise an exception.
        """

        class ConcreteWorker(AbstractWorker):
            pass

        fake_builder = Builder("fake_builder")
        worker = ConcreteWorker("worker", "pass")
        workerforbuilder = AbstractWorkerForBuilder(fake_builder)
        # FIXME: This should call attached, instead of setting the attribute
        # directly
        workerforbuilder.worker = worker

        # The following shouldn't raise an exception.
        workerforbuilder.buildStarted()



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members


from twisted.internet import defer

from buildbot.plugins import schedulers
from buildbot.test.util.integration import RunMasterBase


# This integration test creates a master and worker environment,
# with one builders and a shellcommand step
# meant to be a template for integration steps
class ShellMaster(RunMasterBase):
    def create_config(self):
        c = {}
        from buildbot.config import BuilderConfig
        from buildbot.plugins import steps
        from buildbot.process.factory import BuildFactory

        c['schedulers'] = [
            schedulers.AnyBranchScheduler(name="sched1", builderNames=["testy1"]),
            schedulers.ForceScheduler(name="sched2", builderNames=["testy2"]),
        ]
        f = BuildFactory()
        f.addStep(steps.ShellCommand(command='echo hello'))
        c['builders'] = [
            BuilderConfig(name=name, workernames=["local1"], factory=f)
            for name in ['testy1', 'testy2']
        ]
        return c

    @defer.inlineCallbacks
    def test_shell(self):
        cfg = self.create_config()
        yield self.setup_master(cfg)

        change = {
            "branch": "master",
            "files": ["foo.c"],
            "author": "me@foo.com",
            "committer": "me@foo.com",
            "comments": "good stuff",
            "revision": "HEAD",
            "project": "none",
        }
        # switch the configuration of the scheduler, and make sure the correct builder is run
        cfg['schedulers'] = [
            schedulers.AnyBranchScheduler(name="sched1", builderNames=["testy2"]),
            schedulers.ForceScheduler(name="sched2", builderNames=["testy1"]),
        ]
        yield self.master.reconfig()
        build = yield self.doForceBuild(wantSteps=True, useChange=change, wantLogs=True)
        self.assertEqual(build['buildid'], 1)
        builder = yield self.master.data.get(('builders', build['builderid']))
        self.assertEqual(builder['name'], 'testy2')
